import logging
from datetime import datetime
from typing import Sequence, Tuple, List, Dict, TypeVar, Type, AsyncGenerator, TYPE_CHECKING

from lightbus.api import Api
from lightbus.exceptions import NothingToListenFor
from lightbus.message import RpcMessage, EventMessage, ResultMessage
from lightbus.serializers import ByFieldMessageSerializer, ByFieldMessageDeserializer
from lightbus.utilities.config import make_from_config_structure

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from lightbus.config import Config

T = TypeVar("T")
logger = logging.getLogger(__name__)


class TransportMetaclass(type):
    def __new__(mcs, name, bases, attrs, **kwds):
        cls = super().__new__(mcs, name, bases, attrs)
        if not hasattr(cls, f"{name}Config") and hasattr(cls, "from_config"):
            cls.Config = make_from_config_structure(
                class_name=name, from_config_method=cls.from_config
            )
        return cls


class Transport(metaclass=TransportMetaclass):
    @classmethod
    def from_config(cls: Type[T], config: "Config") -> T:
        return cls()

    async def open(self):
        """Setup transport prior to use


        Can be used for opening connections, initialisation, etc.
        """
        pass

    async def close(self):
        """Cleanup prior to termination

        Can be used for closing connections etc.
        """
        pass


class RpcTransport(Transport):
    """Implement the sending and receiving of RPC calls"""

    async def call_rpc(self, rpc_message: RpcMessage, options: dict):
        """Publish a call to a remote procedure"""
        raise NotImplementedError()

    async def consume_rpcs(self, apis: Sequence[Api]) -> Sequence[RpcMessage]:
        """Consume RPC calls for the given API"""
        raise NotImplementedError()


class ResultTransport(Transport):
    """Implement the send & receiving of results

    """

    async def get_return_path(self, rpc_message: RpcMessage) -> str:
        raise NotImplementedError()

    async def send_result(
        self, rpc_message: RpcMessage, result_message: ResultMessage, return_path: str
    ):
        """Send a result back to the caller

        Args:
            rpc_message (): The original message received from the client
            result_message (): The result message to be sent back to the client
            return_path (str): The string indicating where to send the result.
                As generated by :ref:`get_return_path()`.
        """
        raise NotImplementedError()

    async def receive_result(
        self, rpc_message: RpcMessage, return_path: str, options: dict
    ) -> ResultMessage:
        """Receive the result for the given message

        Args:
            rpc_message (): The original message sent to the worker
            return_path (str): The string indicated where to receive the result.
                As generated by :ref:`get_return_path()`.
            options (dict): Dictionary of options specific to this particular backend
        """
        raise NotImplementedError()


class EventTransport(Transport):
    """ Implement the sending/consumption of events over a given transport.
    """

    def __init__(
        self,
        serializer=ByFieldMessageSerializer(),
        deserializer=ByFieldMessageDeserializer(EventMessage),
    ):
        self.serializer = serializer
        self.deserializer = deserializer

    async def send_event(self, event_message: EventMessage, options: dict):
        """Publish an event"""
        raise NotImplementedError()

    async def consume(
        self, listen_for: List[Tuple[str, str]], listener_name: str, **kwargs
    ) -> AsyncGenerator[List[EventMessage], None]:
        """Consume messages for the given APIs

        Examples:

            Consuming events::

                listen_for = [
                    ('mycompany.auth', 'user_created'),
                    ('mycompany.auth', 'user_updated'),
                ]
                async with event_transport.consume(listen_for) as event_message:
                    print(event_message)

        """
        raise NotImplementedError(
            f"Event transport {self.__class__.__name__} does not support listening for events"
        )

    async def acknowledge(self, *event_messages):
        """Acknowledge that one or more events were successfully processed"""
        pass

    async def history(
        self,
        api_name,
        event_name,
        start: datetime = None,
        stop: datetime = None,
        start_inclusive: bool = True,
    ) -> AsyncGenerator[EventMessage, None]:
        """Return EventMessages for the given api/event names during the (optionally) given date range.

        Should return newest messages first
        """
        raise NotImplementedError(
            f"Event transport {self.__class__.__name__} does not support event history."
        )

    def _sanity_check_listen_for(self, listen_for):
        """Utility method to sanity check the `listen_for` parameter.

        Call at the start of your consume() implementation.
        """
        if not listen_for:
            raise NothingToListenFor(
                "EventTransport.consume() was called without providing anything "
                'to listen for in the "listen_for" argument.'
            )


class SchemaTransport(Transport):
    """ Implement sharing of lightbus API schemas
    """

    async def store(self, api_name: str, schema: Dict, ttl_seconds: int):
        """Store a schema for the given API"""
        raise NotImplementedError()

    async def ping(self, api_name: str, schema: Dict, ttl_seconds: int):
        """Keep alive a schema already stored via store()

        The defaults to simply calling store() on the assumption that this
        will cause the ttl to be updated. Backends may choose to
        customise this logic.
        """
        await self.store(api_name, schema, ttl_seconds)

    async def load(self) -> Dict[str, Dict]:
        """Load the schema for all APIs

        Should return a mapping of API names to schemas
        """
        raise NotImplementedError()



"""Data structures used for specifying which features a lightbus process should provide"""
from enum import Enum


class Feature(Enum):
    RPCS = "rpcs"
    EVENTS = "events"
    TASKS = "tasks"


ALL_FEATURES = tuple(Feature)



import sys
import asyncio
import logging
import threading
from concurrent.futures.thread import ThreadPoolExecutor
from time import time
from typing import Coroutine, TYPE_CHECKING
import datetime

import aioredis

from lightbus.exceptions import CannotBlockHere

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import,cyclic-import
    from schedule import Job

logger = logging.getLogger(__name__)

PYTHON_VERSION = tuple(sys.version_info)


def block(coroutine: Coroutine, loop=None, *, timeout=None):
    """Call asynchronous code synchronously

    Note that this cannot be used inside an event loop.
    """
    loop = loop or get_event_loop()
    if loop.is_running():
        if hasattr(coroutine, "close"):
            coroutine.close()
        raise CannotBlockHere(
            "It appears you have tried to use a blocking API method "
            "from within an event loop. Unfortunately this is unsupported. "
            "Instead, use the async version of the method."
        )
    try:
        if timeout is None:
            val = loop.run_until_complete(coroutine)
        else:
            val = loop.run_until_complete(asyncio.wait_for(coroutine, timeout=timeout))
    except Exception as e:
        # The intention here is to get sensible stack traces from exceptions within blocking calls
        raise e
    return val


def get_event_loop():
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


async def cancel(*tasks):
    """Useful for cleaning up tasks in tests"""
    # pylint: disable=broad-except
    ex = None
    for task in tasks:
        if task is None:
            continue

        # Cancel all the tasks any pull out any exceptions
        if not task.cancelled():
            task.cancel()
        try:
            await task
            task.result()
        except asyncio.CancelledError:
            pass
        except Exception as e:
            # aioredis hides errors which occurred within a pipline
            # by wrapping them up in a PipelineError. So see if that is
            # the case here
            if (
                # An aioredis pipeline error
                isinstance(e, aioredis.PipelineError)
                # Where one of the errors that caused it was a CancelledError
                and len(e.args) > 1
                and asyncio.CancelledError in map(type, e.args[1])
            ):
                pass
            else:
                # If there was an exception, and this is the first
                # exception we've seen, then stash it away for later
                if ex is None:
                    ex = e

    # Now raise the first exception we saw, if any
    if ex:
        raise ex


async def cancel_and_log_exceptions(*tasks):
    """Cancel tasks and log any exceptions

    This is useful when shutting down, when tasks need to be cancelled any anything
    that goes wrong should be logged but will not otherwise be dealt with.
    """
    for task in tasks:
        try:
            await cancel(task)
        except Exception as e:
            # pylint: disable=broad-except
            logger.exception(e)
            logger.error(
                "Error encountered when shutting down task %s. Exception logged, will now move on.",
                task,
            )


async def run_user_provided_callable(callable_, args, kwargs):
    """Run user provided code

    If the callable is blocking (i.e. a regular function) it will be
    moved to its own thread and awaited. The purpose of this thread is to:

        1. Allow other tasks to continue as normal
        2. Allow user code to start its own event loop where needed

    If an async function is provided it will be awaited as-is, and no
    child thread will be created.

    The callable will be called with the given args and kwargs
    """
    if asyncio.iscoroutinefunction(callable_):
        try:
            return await callable_(*args, **kwargs)
        except Exception as e:
            exception = e
    else:
        try:
            thread_pool_executor = ThreadPoolExecutor(
                thread_name_prefix="user_provided_callable_tpe"
            )
            future = asyncio.get_event_loop().run_in_executor(
                executor=thread_pool_executor, func=lambda: callable_(*args, **kwargs)
            )
            return await future
        except Exception as e:
            exception = e

    logger.debug(f"Error in user provided callable: {repr(exception)}")
    raise exception


async def call_every(*, callback, timedelta: datetime.timedelta, also_run_immediately: bool):
    """Call callback every timedelta

    If also_run_immediately is set then the callback will be called before any waiting
    happens. If the callback's result is awaitable then it will be awaited

    Callback execution time is accounted for in the scheduling. If the execution takes
    2 seconds, and timedelta is 10 seconds, then call_every() will wait 8 seconds
    before the subsequent execution.
    """
    first_run = True
    while True:
        start_time = time()
        if not first_run or also_run_immediately:
            await run_user_provided_callable(callback, args=[], kwargs={})
        total_execution_time = time() - start_time
        sleep_time = max(0.0, timedelta.total_seconds() - total_execution_time)
        await asyncio.sleep(sleep_time)
        first_run = False


async def call_on_schedule(callback, schedule: "Job", also_run_immediately: bool):
    first_run = True
    while True:
        schedule._schedule_next_run()

        if not first_run or also_run_immediately:
            schedule.last_run = datetime.datetime.now()
            await run_user_provided_callable(callback, args=[], kwargs={})

        td = schedule.next_run - datetime.datetime.now()
        await asyncio.sleep(td.total_seconds())
        first_run = False



import asyncio
import logging
from datetime import timedelta
from functools import wraps
from inspect import iscoroutinefunction
from typing import List, Tuple, Coroutine, Union, Sequence, TYPE_CHECKING, Callable

from lightbus.client.utilities import queue_exception_checker, Error, ErrorQueueType, OnError
from lightbus.exceptions import (
    InvalidSchedule,
    BusAlreadyClosed,
    UnsupportedUse,
    AsyncFunctionOrMethodRequired,
)
from lightbus.log import LBullets
from lightbus.utilities.async_tools import (
    block,
    get_event_loop,
    cancel,
    call_every,
    call_on_schedule,
    cancel_and_log_exceptions,
)
from lightbus.utilities.features import Feature, ALL_FEATURES
from lightbus.utilities.frozendict import frozendict

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from schedule import Job
    from lightbus.config import Config
    from lightbus.client.subclients.event import EventClient
    from lightbus.client.subclients.rpc_result import RpcResultClient
    from lightbus.plugins import PluginRegistry
    from lightbus.schema import Schema
    from lightbus.api import Api, ApiRegistry
    from lightbus.transports.registry import TransportRegistry
    from lightbus.hooks import HookRegistry

__all__ = ["BusClient"]


logger = logging.getLogger(__name__)


def raise_queued_errors(fn):
    """Decorator to raise any errors which have appeared on the bus' error_queue during execution

    There are two situations in which we can categorise errors:

        1. Errors which occur from a user directly accessing the bus.
           For example, calling an RPC or firing an event
        2. Errors which occur within background tasks which are running as
           part of the Lightbus worker

    This decorator handles the former. The BusClient's error_monitor() handles the second
    """
    if not iscoroutinefunction(fn):
        raise AsyncFunctionOrMethodRequired(
            "@raise_queued_errors can only be used on async methods"
        )

    @wraps(fn)
    async def _wrapped(self: "BusClient", *args, **kwargs):
        result = await fn(self, *args, **kwargs)
        await self.raise_errors()
        return result

    return _wrapped


class BusClient:
    """Provides a the lower level interface for accessing the bus

    The low-level `BusClient` is less expressive than the interface provided by `BusPath`,
    but does allow for more control in some situations.

    All functionality in `BusPath` is provided by `BusClient`.
    """

    def __init__(
        self,
        config: "Config",
        schema: "Schema",
        plugin_registry: "PluginRegistry",
        hook_registry: "HookRegistry",
        event_client: "EventClient",
        rpc_result_client: "RpcResultClient",
        api_registry: "ApiRegistry",
        transport_registry: "TransportRegistry",
        error_queue: ErrorQueueType,
        features: Sequence[Union[Feature, str]] = ALL_FEATURES,
    ):
        self.config = config
        self.schema = schema
        self.plugin_registry = plugin_registry
        self.hook_registry = hook_registry
        self.event_client = event_client
        self.rpc_result_client = rpc_result_client
        self.api_registry = api_registry
        # The transport registry isn't actually used by the bus client, but it is
        # useful to have it available as a property on the client.
        self.transport_registry = transport_registry
        self.error_queue = error_queue

        # Coroutines added via schedule/every/add_background_task which should be started up
        # once the worker starts
        self._background_coroutines = []
        # Tasks produced from the values in self._background_coroutines. Will be closed on bus shutdown
        self._background_tasks = []

        self.features: List[Union[Feature, str]] = []
        self.set_features(ALL_FEATURES)

        self.exit_code = 0
        self._closed = False
        self._worker_tasks = set()
        self._lazy_load_complete = False

        # Used to detect if the event monitor is running
        self._error_monitor_lock = asyncio.Lock()

    def close(self):
        """Close the bus client

        This will cancel all tasks and close all transports/connections
        """
        block(self.close_async())

    async def close_async(self):
        """Async version of close()"""
        if self._closed:
            raise BusAlreadyClosed()

        await cancel_and_log_exceptions(*self._background_tasks)

        await self.event_client.close()
        await self.rpc_result_client.close()
        await self.schema.close()

        while not self.error_queue.empty():
            logger.error(self.error_queue.get_nowait())

        self._closed = True

    @property
    def loop(self):
        return get_event_loop()

    def run_forever(self):
        block(self.start_worker())

        self._actually_run_forever()
        logger.debug("Main thread event loop was stopped")

        # Close down the worker
        logger.debug("Stopping worker")
        block(self.stop_worker())

        # Close down the client
        logger.debug("Closing bus")
        block(self.close_async())

        return self.exit_code

    async def start_worker(self):
        """Worker startup procedure"""
        # Ensure an event loop exists
        get_event_loop()

        self._worker_tasks = set()

        # Start monitoring for errors on the error queue
        error_monitor_task = asyncio.ensure_future(self.error_monitor())
        self._error_monitor_task = error_monitor_task
        self._worker_tasks.add(self._error_monitor_task)

        # Features setup & logging
        if not self.api_registry.all() and Feature.RPCS in self.features:
            logger.info("Disabling serving of RPCs as no APIs have been registered")
            self.features.remove(Feature.RPCS)

        logger.info(
            LBullets(
                f"Enabled features ({len(self.features)})", items=[f.value for f in self.features]
            )
        )

        disabled_features = set(ALL_FEATURES) - set(self.features)
        logger.info(
            LBullets(
                f"Disabled features ({len(disabled_features)})",
                items=[f.value for f in disabled_features],
            )
        )

        # Api logging
        logger.info(
            LBullets(
                "APIs in registry ({})".format(len(self.api_registry.all())),
                items=self.api_registry.names(),
            )
        )

        # Push all registered APIs into the global schema
        for api in self.api_registry.all():
            await self.schema.add_api(api)

        # We're running as a worker now (e.g. lightbus run), so
        # do the lazy loading immediately
        await self.lazy_load_now()

        # Setup schema monitoring
        monitor_task = asyncio.ensure_future(
            queue_exception_checker(self.schema.monitor(), self.error_queue)
        )

        logger.info("Executing before_worker_start & on_start hooks...")
        await self.hook_registry.execute("before_worker_start")
        logger.info("Execution of before_worker_start & on_start hooks was successful")

        # Setup RPC consumption
        if Feature.RPCS in self.features:
            consume_rpc_task = asyncio.ensure_future(
                queue_exception_checker(self.consume_rpcs(), self.error_queue)
            )
        else:
            consume_rpc_task = None

        # Start off any registered event listeners
        if Feature.EVENTS in self.features:
            await self.event_client.start_registered_listeners()

        # Start off any background tasks
        if Feature.TASKS in self.features:
            for coroutine in self._background_coroutines:
                task = asyncio.ensure_future(queue_exception_checker(coroutine, self.error_queue))
                self._background_tasks.append(task)

        self._worker_tasks.add(consume_rpc_task)
        self._worker_tasks.add(monitor_task)

    async def stop_worker(self):
        logger.debug("Stopping worker")

        # Cancel the tasks we created above
        await cancel(*self._worker_tasks)

        logger.info("Executing after_worker_stopped & on_stop hooks...")
        await self.hook_registry.execute("after_worker_stopped")
        logger.info("Execution of after_worker_stopped & on_stop hooks was successful")

    def _actually_run_forever(self):  # pragma: no cover
        """Simply start the loop running forever

        This just makes testing easier as we can mock out this method
        """
        self.loop.run_forever()

    async def raise_errors(self):
        # If the error monitor is running then just return, as that means we are
        # running as a worker and so can rely on the error monitor to pickup the
        # errors which an happen in the various background tasks
        if self._error_monitor_lock.locked():
            return

        # Check queue for errors
        try:
            error: Error = self.error_queue.get_nowait()
        except asyncio.QueueEmpty:
            # No errors, everything ok
            return
        else:
            # If there is an error then raise it
            raise error.value

    async def error_monitor(self):
        async with self._error_monitor_lock:
            error: Error = await self.error_queue.get()
            logger.debug(f"Bus client event monitor detected an error, will shutdown.")

            if error.should_show_error():
                logger.error(str(error))
                if error.help:
                    logger.error(error.help)
            else:
                logger.info(error.value)

            self.exit_code = error.exit_code
            self.stop_loop()

    def stop_loop(self):
        logger.debug("Stopping event loop")
        self.loop.stop()

    def request_shutdown(self):
        logger.debug("Requesting Lightbus shutdown")
        self.error_queue.put_nowait(
            Error(
                type=KeyboardInterrupt,
                value=KeyboardInterrupt("Shutdown requested"),
                traceback=None,
                invoking_stack=None,
                exit_code=0,
            )
        )

    async def lazy_load_now(self):
        """Perform lazy tasks immediately

        When lightbus is used as a client it performs network tasks
        lazily. This speeds up import of your bus module, and prevents
        getting surprising errors at import time.

        However, in some cases you may wish to hurry up these lazy tasks
        (or perform them at a known point). In which case you can call this
        method to execute them immediately.
        """
        if self._lazy_load_complete:
            return

        # 1. Load the schema
        logger.debug("Loading schema...")
        await self.schema.ensure_loaded_from_bus()

        logger.info(
            LBullets(
                "Loaded the following remote schemas ({})".format(len(self.schema.remote_schemas)),
                items=self.schema.remote_schemas.keys(),
            )
        )

        # 2. Ensure the schema transport is loaded (other transports will be
        #    loaded as the need arises, but we need schema information from the get-go)
        await self.schema.ensure_loaded_from_bus()

        # 3. Add any local APIs to the schema
        for api in self.api_registry.all():
            await self.schema.add_api(api)

        logger.info(
            LBullets(
                "Loaded the following local schemas ({})".format(len(self.schema.remote_schemas)),
                items=self.schema.local_schemas.keys(),
            )
        )

        # 4. Done
        self._lazy_load_complete = True

    # RPCs

    @raise_queued_errors
    async def consume_rpcs(self, apis: List["Api"] = None):
        """Start a background task to consume RPCs

        This will consumer RPCs on APIs which have been registered with this
        bus client.
        """
        return await self.rpc_result_client.consume_rpcs(apis)

    @raise_queued_errors
    async def call_rpc_remote(
        self, api_name: str, name: str, kwargs: dict = frozendict(), options: dict = frozendict()
    ):
        """Perform an RPC call

        Call an RPC and return the result.
        """
        await self.lazy_load_now()
        return await self.rpc_result_client.call_rpc_remote(
            api_name=api_name, name=name, kwargs=kwargs, options=options
        )

    # Events

    @raise_queued_errors
    async def fire_event(self, api_name, name, kwargs: dict = None, options: dict = None):
        """Fire an event onto the bus"""
        await self.lazy_load_now()
        return await self.event_client.fire_event(
            api_name=api_name, name=name, kwargs=kwargs, options=options
        )

    def listen_for_event(
        self,
        api_name: str,
        name: str,
        listener: Callable,
        listener_name: str,
        options: dict = None,
        on_error: OnError = OnError.SHUTDOWN,
    ):
        """Listen for a single event

        Wraps `listen_for_events()`
        """
        self.listen_for_events(
            [(api_name, name)], listener, listener_name=listener_name, options=options or {}
        )

    def listen_for_events(
        self,
        events: List[Tuple[str, str]],
        listener: Callable,
        listener_name: str,
        options: dict = None,
        on_error: OnError = OnError.SHUTDOWN,
    ):
        """Listen for a list of events

        `events` is in the form:

            events=[
                ('company.first_api', 'event_name'),
                ('company.second_api', 'event_name'),
            ]

        `listener_name` is an arbitrary string which uniquely identifies this listener.
        This can generally be the same as the function name of the `listener` callable, but
        it should not change once deployed.
        """
        return self.event_client.listen(
            events=events,
            listener=listener,
            listener_name=listener_name,
            options=options,
            on_error=on_error,
        )

    def add_background_task(self, coroutine: Union[Coroutine, asyncio.Future]):
        """Run a coroutine in the background

        The provided coroutine will be run in the background once
        Lightbus startup is complete.

        The coroutine will be cancelled when the bus client is closed.

        The Lightbus process will exit if the coroutine raises an exception.
        See lightbus.utilities.async_tools.check_for_exception() for details.
        """

        # Store coroutine for starting once the worker starts
        self._background_coroutines.append(coroutine)

    # Utilities

    def set_features(self, features: List[Union[Feature, str]]):
        """Set the features this bus clients should serve.

        Features should be a list of: `rpcs`, `events`, `tasks`
        """
        features = list(features)
        for i, feature in enumerate(features):
            try:
                features[i] = Feature(feature)
            except ValueError:
                features_str = ", ".join([f.value for f in Feature])
                raise UnsupportedUse(f"Feature '{feature}' is not one of: {features_str}\n")

        self.features = features

    # Hooks

    def _make_hook_decorator(self, name, before_plugins=False, callback=None):
        if callback and not callable(callback):
            raise AssertionError("The provided callback is not callable")

        if callback:
            self.hook_registry.register_callback(name, callback, before_plugins)
            return None
        else:

            def hook_decorator(fn):
                self.hook_registry.register_callback(name, fn, before_plugins)
                return fn

            return hook_decorator

    def on_start(self, callback=None, *, before_plugins=False):
        """Decorator to register a function to be called before the worker starts up

        Callback will be called with the following arguments:

            callback(self, *, client: "BusClient")
        """
        return self.before_worker_start(callback, before_plugins=before_plugins)

    def on_stop(self, callback=None, *, before_plugins=False):
        """Decorator to register a function to be called after the worker stops

        Callback will be called with the following arguments:

            callback(self, *, client: "BusClient")
        """
        return self.before_worker_start(callback, before_plugins=before_plugins)

    def before_worker_start(self, callback=None, *, before_plugins=False):
        """Decorator to register a function to be called before the worker starts up

        See `on_start()`
        """
        return self._make_hook_decorator("before_worker_start", before_plugins, callback)

    def after_worker_stopped(self, callback=None, *, before_plugins=False):
        """Decorator to register a function to be called after the worker stops

        See `on_stop()`
        """
        return self._make_hook_decorator("after_worker_stopped", before_plugins, callback)

    def before_rpc_call(self, callback=None, *, before_plugins=False):
        """Decorator to register a function to be called prior to an RPC call

        Callback will be called with the following arguments:

            callback(self, *, rpc_message: RpcMessage, client: "BusClient")
        """
        return self._make_hook_decorator("before_rpc_call", before_plugins, callback)

    def after_rpc_call(self, callback=None, *, before_plugins=False):
        """Decorator to register a function to be called after an RPC call

        Callback will be called with the following arguments:

            callback(self, *, rpc_message: RpcMessage, result_message: ResultMessage, client: "BusClient")
        """
        return self._make_hook_decorator("after_rpc_call", before_plugins, callback)

    def before_rpc_execution(self, callback=None, *, before_plugins=False):
        """Decorator to register a function to be called prior to a local RPC execution

        Callback will be called with the following arguments:

            callback(self, *, rpc_message: RpcMessage, client: "BusClient")
        """
        return self._make_hook_decorator("before_rpc_execution", before_plugins, callback)

    def after_rpc_execution(self, callback=None, *, before_plugins=False):
        """Decorator to register a function to be called after a local RPC execution

        Callback will be called with the following arguments:

            callback(self, *, rpc_message: RpcMessage, result_message: ResultMessage, client: "BusClient")
        """
        return self._make_hook_decorator("after_rpc_execution", before_plugins, callback)

    def before_event_sent(self, callback=None, *, before_plugins=False):
        """Decorator to register a function to be called prior to an event being sent

        Callback will be called with the following arguments:

            callback(self, *, event_message: EventMessage, client: "BusClient")
        """
        return self._make_hook_decorator("before_event_sent", before_plugins, callback)

    def after_event_sent(self, callback=None, *, before_plugins=False):
        """Decorator to register a function to be called after an event was sent

        Callback will be called with the following arguments:

            callback(self, *, event_message: EventMessage, client: "BusClient")
        """
        return self._make_hook_decorator("after_event_sent", before_plugins, callback)

    def before_event_execution(self, callback=None, *, before_plugins=False):
        """Decorator to register a function to be called prior to a local event handler execution

        Callback will be called with the following arguments:

            callback(self, *, event_message: EventMessage, client: "BusClient")
        """
        return self._make_hook_decorator("before_event_execution", before_plugins, callback)

    def after_event_execution(self, callback=None, *, before_plugins=False):
        """Decorator to register a function to be called after a local event handler execution

        Callback will be called with the following arguments:

            callback(self, *, event_message: EventMessage, client: "BusClient")
        """
        return self._make_hook_decorator("after_event_execution", before_plugins, callback)

    # Scheduling

    def every(
        self,
        *,
        seconds=0,
        minutes=0,
        hours=0,
        days=0,
        also_run_immediately=False,
        **timedelta_extra,
    ):
        """Call a coroutine at the specified interval

        This is a simple scheduling mechanism which you can use in your bus module to setup
        recurring tasks. For example:

            bus = lightbus.create()

            @bus.client.every(seconds=30)
            def my_func():
                print("Hello")

        This can also be used to decorate async functions. In this case the function will be awaited.

        Note that the timing is best effort and is not guaranteed. That being said, execution
        time is accounted for.

        See Also:

            @bus.client.schedule()
        """
        td = timedelta(seconds=seconds, minutes=minutes, hours=hours, days=days, **timedelta_extra)

        if td.total_seconds() == 0:
            raise InvalidSchedule(
                "The @bus.client.every() decorator must be provided with a non-zero time argument."
                " Ensure you are passing at least one time argument, and that it has a non-zero"
                " value."
            )

        # TODO: There is an argument that the backgrounding of this should be done only after
        #       on_start() has been fired. Otherwise this will be run before the on_start() setup
        #       has happened in cases where also_run_immediately=True.
        def wrapper(f):
            coroutine = call_every(  # pylint: assignment-from-no-return
                callback=f, timedelta=td, also_run_immediately=also_run_immediately
            )
            self.add_background_task(coroutine)
            return f

        return wrapper

    def schedule(self, schedule: "Job", also_run_immediately=False):
        """Call a coroutine on the specified schedule

        Schedule a task using the `schedule` library:

            import lightbus
            import schedule

            bus = lightbus.create()

            # Run the task every 1-3 seconds, varying randomly
            @bus.client.schedule(schedule.every(1).to(3).seconds)
            def do_it():
                print("Hello using schedule library")

        This can also be used to decorate async functions. In this case the function will be awaited.

        See Also:

            @bus.client.every()
        """

        def wrapper(f):
            coroutine = call_on_schedule(
                callback=f, schedule=schedule, also_run_immediately=also_run_immediately
            )
            self.add_background_task(coroutine)
            return f

        return wrapper

    # API registration

    def register_api(self, api: "Api"):
        """Register an API with this bus client

        You must register APIs which you wish to fire events
        on or handle RPCs calls for.

        See Also: https://lightbus.org/explanation/apis/
        """
        self.api_registry.add(api)



""" Serializers suitable for transports which support multiple fields per message

These serializers handle moving data to/from a dictionary
format. The format looks like this::

    # Message metadata first. Each value is implicitly a utf8 string
    id: 'ZOCTLh1CEeimW3gxwcOTbg=='
    api_name: 'my_company.auth'
    procedure_name: 'check_password'
    return_path: 'redis+key://my_company.auth.check_password:result:ZOCTLh1CEeimW3gxwcOTbg=='

    # kwargs follow, each encoded with the provided encoder (in this case JSON)
    kw:username: '"admin"'
    kw:password: '"secret"'

"""
from typing import TYPE_CHECKING

from lightbus.serializers.base import (
    decode_bytes,
    sanity_check_metadata,
    MessageSerializer,
    MessageDeserializer,
)

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from lightbus import Message


class ByFieldMessageSerializer(MessageSerializer):
    def __call__(self, message: "Message") -> dict:
        """Takes a message object and returns a serialised dictionary representation

        See the module-level docs (above) for further details
        """
        serialized = message.get_metadata()
        for k, v in message.get_kwargs().items():
            serialized[":{}".format(k)] = self.encoder(v)
        return serialized


class ByFieldMessageDeserializer(MessageDeserializer):
    def __call__(self, serialized: dict, *, native_id=None, **extra):
        """Takes a dictionary of serialised fields and returns a Message object

        See the module-level docs (above) for further details
        """
        metadata = {}
        kwargs = {}

        for k, v in serialized.items():
            k = decode_bytes(k)
            v = decode_bytes(v)

            if not k:
                continue

            # kwarg fields start with a ':', everything else is metadata
            if k[0] == ":":
                # kwarg values need decoding
                kwargs[k[1:]] = self.decoder(v)
            else:
                # metadata args are implicitly strings, so we don't need to decode them
                metadata[k] = v

        sanity_check_metadata(self.message_class, metadata)

        if "native_id" in metadata:
            native_id = metadata.pop("native_id")

        return self.message_class.from_dict(
            metadata=metadata, kwargs=kwargs, native_id=native_id, **extra
        )



import logging

from typing import Tuple, Any

import lightbus.creation
from lightbus import configure_logging, BusPath
import lightbus.client
from lightbus.config import Config
from lightbus.exceptions import NoBusFoundInBusModule

logger = logging.getLogger(__name__)


def import_bus(args) -> Tuple[Any, BusPath]:
    bus_module = lightbus.creation.import_bus_module(args.bus_module_name)
    try:
        return bus_module, bus_module.bus
    except AttributeError:
        raise NoBusFoundInBusModule(
            f"Bus module at {bus_module.__file__} contains no variable named 'bus'. "
            f"Your bus module should contain the line 'bus = lightbus.create()'."
        )


def setup_logging(override: str, config: Config):
    configure_logging(log_level=(override or config.bus().log_level.value).upper())


def setup_common_arguments(parser):
    """Set common arguments needed by all commands"""
    general_argument_group = parser.add_argument_group(title="Common arguments")
    general_argument_group.add_argument(
        "--bus",
        "-b",
        dest="bus_module_name",
        metavar="BUS_MODULE",
        help=(
            "The bus module to import. Example 'bus', 'my_project.bus'. Defaults to "
            "the value of the LIGHTBUS_MODULE environment variable, or 'bus'"
        ),
    )
    general_argument_group.add_argument(
        "--service-name",
        "-s",
        help="Name of service in which this process resides. YOU SHOULD "
        "LIKELY SET THIS IN PRODUCTION. Can also be set using the "
        "LIGHTBUS_SERVICE_NAME environment. Will default to a random string.",
    )
    general_argument_group.add_argument(
        "--process-name",
        "-p",
        help="A unique name of this process within the service. Can also be set using the "
        "LIGHTBUS_PROCESS_NAME environment. Will default to a random string.",
    )
    general_argument_group.add_argument(
        "--config", dest="config_file", help="Config file to load, JSON or YAML", metavar="FILE"
    )
    general_argument_group.add_argument(
        "--log-level",
        help="Set the log level. Overrides any value set in config. "
        "One of debug, info, warning, critical, exception.",
        metavar="LOG_LEVEL",
    )



import asyncio
import logging
import sys
from enum import Enum
from functools import wraps
from traceback import extract_stack, StackSummary, format_exception, format_list

from types import TracebackType
from typing import Coroutine, NamedTuple, Type, TYPE_CHECKING, Optional

from lightbus.exceptions import InvalidName
from lightbus.utilities.internal_queue import InternalQueue

logger = logging.getLogger(__name__)


class OnError(Enum):
    SHUTDOWN = "shutdown"
    ACKNOWLEDGE_AND_LOG = "log"


ErrorQueueType = InternalQueue["Error"]


class Error(NamedTuple):
    """An error within a coroutine.

    Specifically, this is typically used to represent errors which
    occurred in child tasks.

    This stores both the data from sys.exc_info(), as well as an additional
    `invoking_stack` value. This `invoking_stack` stores the stack at the point
    where the coroutine was created, whereas the error's traceback is typically local
    to a child asyncio task.

    As a result, we can display a more useful error message. This error message
    can show both the exception and traceback for the error, and also the tracback
    for the location where the coroutine was created.
    """

    type: Type[BaseException]
    value: BaseException
    traceback: Optional[TracebackType]
    invoking_stack: Optional[StackSummary]
    help: str = ""
    exit_code: int = 1

    @property
    def invoking_traceback_str(self) -> str:
        if self.invoking_stack:
            return "".join(format_list(self.invoking_stack)).strip()
        else:
            return ""

    @property
    def exception_str(self) -> str:
        if self.type and self.value and self.traceback:
            return "".join(format_exception(self.type, self.value, self.traceback)).strip()
        else:
            return ""

    def __str__(self):
        return (
            f"A coroutine raised an error, it was invoked at:\n\n"
            f"{self.invoking_traceback_str or 'Not available'}\n\n"
            f"The exception was:\n\n"
            f"{self.exception_str or 'Not available'}"
        )

    def should_show_error(self):
        """Should this error be logged out?"""
        return self.exit_code != 0


def validate_event_or_rpc_name(api_name: str, type_: str, name: str):
    """Validate that the given RPC/event name is ok to use"""
    if not name:
        raise InvalidName(f"Empty {type_} name specified when calling API {api_name}")

    if name.startswith("_"):
        raise InvalidName(
            f"You can not use '{api_name}.{name}' as an {type_} because it starts with an underscore. "
            f"API attributes starting with underscores are not available on the bus."
        )


def queue_exception_checker(coroutine: Coroutine, error_queue: InternalQueue, help: str = ""):
    """Await a coroutine and place any errors into the provided queue

    Items added to the queue will be of the `Error` type. This `Error` type
    allows for more informative error messages.

    See Also: Error
    """
    invoking_stack = extract_stack()[:-1]

    @wraps(coroutine)
    async def coroutine_():
        try:
            await coroutine
        except asyncio.CancelledError:
            # Was cancelled by the caller, so just re-raise the cancelled error and
            # let the coroutine finish as requested
            raise
        except Exception as e:
            # This is an exception which is going to need to be dealt with

            # We set a flag on each exception we add to the queue to indicate
            # that it has been enqueued. This prevents an exception being added to the
            # queue multiple times as it propagates up potentially multiple levels
            # of asyncio tasks (which of which would presumably be wrapped up in queue_exception_checker())
            if not getattr(e, "enqueued", False):
                # Flag not set, so add this error to the queue, and set the flag on it
                e.enqueued = True
                error = Error(*sys.exc_info(), invoking_stack, help)
                await error_queue.put(error)

            raise

    return coroutine_()



import asyncio
import logging
from argparse import ArgumentParser, _ArgumentGroup, Namespace
from typing import Dict, Type, TypeVar, NamedTuple, TYPE_CHECKING

from collections import OrderedDict

from lightbus.schema.schema import Parameter
from lightbus.exceptions import PluginHookNotFound, LightbusShutdownInProgress
from lightbus.message import RpcMessage, EventMessage, ResultMessage
from lightbus.utilities.config import make_from_config_structure
from lightbus.utilities.importing import load_entrypoint_classes

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from lightbus.config import Config
    from lightbus.client import BusClient

ENTRYPOINT_NAME = "lightbus_plugins"


logger = logging.getLogger(__name__)

T = TypeVar("T")


class PluginMetaclass(type):
    def __new__(mcs, name, bases, attrs, **kwds):
        cls = super().__new__(mcs, name, bases, attrs)
        if not hasattr(cls, f"{name}Config") and hasattr(cls, "from_config"):
            cls.Config = make_from_config_structure(
                class_name=name,
                from_config_method=cls.from_config,
                extra_parameters=[Parameter("enabled", bool, default=False)],
            )
        return cls


class LightbusPlugin(metaclass=PluginMetaclass):
    priority = 1000

    @classmethod
    def from_config(cls: Type[T], *, config) -> T:
        return cls()

    def __str__(self):
        return "{}.{}".format(self.__class__.__module__, self.__class__.__name__)

    async def before_parse_args(self, *, parser: ArgumentParser, subparsers: _ArgumentGroup):
        """ Setup command line argument parser

        Configuration is not available within the before_parse_args()
        hook. This hooks will be called in a separate throw-away instance of this Plugin.

        Note that we don't have an after_parse_args plugin hook. Instead we use the receive_args
        hook which is called once we have instantiated our plugins.
        """
        pass

    async def receive_args(self, args: Namespace):
        """ Received the parsed command line arguments

        This will be called once the plugins have been instantiated. You can therefore
        safely modify your plugins state at this time (unlike in before_parse_args)
        """
        pass

    async def before_worker_start(self, *, client: "BusClient"):
        """Called after the worker process is setup, but before any RPC or Event consumption"""
        pass

    async def after_worker_stopped(self, *, client: "BusClient"):
        """Called during worker process shutdown (after all RPC and Event consumption has been stopped)"""
        pass

    async def before_rpc_call(self, *, rpc_message: RpcMessage, client: "BusClient"):
        """Called before `rpc_message` is sent to the bus

        `rpc_message` is mutable and may therefore be changed prior to being sent
        """
        pass

    async def after_rpc_call(
        self, *, rpc_message: RpcMessage, result_message: ResultMessage, client: "BusClient"
    ):
        """Called after an RPC response has been received back from the bus

        `result_message` is mutable and may therefore be changed prior to it being
        passed back to the calling code
        """
        pass

    async def before_rpc_execution(self, *, rpc_message: RpcMessage, client: "BusClient"):
        """Called once an incoming RPC call is received, but before it is executed

        `rpc_message` is mutable and may therefore be changed prior to being passed to the
        handling code.
        """
        pass

    async def after_rpc_execution(
        self, *, rpc_message: RpcMessage, result_message: ResultMessage, client: "BusClient"
    ):
        """Called after an incoming RPC call has been received and executed

        `result_message` is mutable and may therefore be changed prior to being
        sent on the bus back to the caller.
        """

    async def before_event_sent(self, *, event_message: EventMessage, client: "BusClient"):
        """Called before an event to sent onto the bus

        `event_message` is mutable and may therefore be changed prior to being sent
        """
        pass

    async def after_event_sent(self, *, event_message: EventMessage, client: "BusClient"):
        """Called after an event has been sent onto the bus"""
        pass

    async def before_event_execution(self, *, event_message: EventMessage, client: "BusClient"):
        """Called once an incoming event call is received, but before its handler is executed

        `event_message` is mutable and may therefore be changed prior to being passed to the
        handling code.
        """
        pass

    async def after_event_execution(self, *, event_message: EventMessage, client: "BusClient"):
        """Called after an incoming event call has been received and its handler executed"""

    async def exception(self, *, e: Exception):
        """Called during the handling of Exception e

        This hook will only be called when running within a lightbus worker (ie. via `lightbus run`).
        """
        pass


def instantiate_plugin(config: "Config", plugin_config: NamedTuple, cls: Type[LightbusPlugin]):
    options = plugin_config._asdict()
    options.pop("enabled")
    return cls.from_config(config=config, **options)


def find_plugins() -> Dict[str, Type[LightbusPlugin]]:
    """Discover available plugin classes using the 'lightbus_plugins' entrypoint
    """
    available_plugin_classes = load_entrypoint_classes(ENTRYPOINT_NAME)
    available_plugin_classes = sorted(available_plugin_classes, key=lambda v: v[-1].priority)

    plugins = OrderedDict()
    for module_name, name, plugin in available_plugin_classes:
        if name in plugins:
            pass
        plugins[name] = plugin

    return plugins


class PluginRegistry:
    VALID_HOOK_NAMES = {k for k in LightbusPlugin.__dict__ if not k.startswith("_")}

    def __init__(self):
        self._plugins = []

    def autoload_plugins(self, config: "Config"):
        """Autoload this registry with plugins from the 'lightbus_plugins' entrypoint"""
        for name, cls in find_plugins().items():
            plugin_config = config.plugin(name)
            if plugin_config.enabled:
                self._plugins.append(
                    instantiate_plugin(config=config, plugin_config=plugin_config, cls=cls)
                )

        return self._plugins

    def set_plugins(self, plugins: list):
        """Manually set the plugins in this registry"""
        self._plugins = plugins

    def is_plugin_loaded(self, plugin_class: Type[LightbusPlugin]):
        return plugin_class in [type(p) for p in self._plugins]

    async def execute_hook(self, name, **kwargs):
        if name not in self.VALID_HOOK_NAMES:
            raise PluginHookNotFound(
                "Plugin hook '{}' could not be found. Must be one of: {}".format(
                    name, ", ".join(self.VALID_HOOK_NAMES)
                )
            )

        return_values = []

        for plugin in self._plugins:
            handler = getattr(plugin, name, None)
            if handler:
                try:
                    return_values.append(await handler(**kwargs))
                except asyncio.CancelledError:
                    raise
                except LightbusShutdownInProgress as e:
                    logger.info("Shutdown in progress: {}".format(e))
                except Exception:
                    raise

        return return_values



import json as jsonlib
import os
from pathlib import Path
from typing import Dict, NamedTuple, Union, TYPE_CHECKING
import urllib.request

import jsonschema
import yaml as yamllib

from lightbus.exceptions import UnexpectedConfigurationFormat
from lightbus.schema.hints_to_schema import python_type_to_json_schemas, SCHEMA_URI
from lightbus.utilities.casting import cast_to_hint
from lightbus.utilities.deforming import deform_to_bus

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from .structure import RootConfig, BusConfig, ApiConfig


class Config:
    """Provides access to configuration options

    There are two forms of configuration:

        * Bus-level configuration, `config.bus()`
        * API-level configuration, `config.api(api_name)`

    Bus-level configuration is global to lightbus. API-level configuration
    will normally have a default catch-all definition, but can be customised
    on a per-api basis.
    """

    _config: "RootConfig"

    def __init__(self, root_config: "RootConfig"):
        self._config = root_config

    def bus(self) -> "BusConfig":
        return self._config.bus

    def api(self, api_name=None) -> "ApiConfig":
        """Returns config for the given API

        If there is no API-specific config available for the
        given api_name, then the root API config will be returned.
        """
        return self._config.apis.get(api_name, None) or self._config.apis["default"]

    def apis(self) -> Dict[str, "ApiConfig"]:
        return self._config.apis

    def plugin(self, plugin_name) -> NamedTuple:
        return getattr(self._config.plugins, plugin_name)

    @classmethod
    def load_file(cls, file_path: Union[str, Path]):
        """Instantiate the config from the given file path

        Files ending in `.json` will be parsed as JSON, otherwise the
        file will be parsed as YAML.
        """

        if str(file_path).startswith("http://") or str(file_path).startswith("https://"):
            response = urllib.request.urlopen(file_path, timeout=5)
            encoded_config = response.read()
            if "json" in response.headers.get("Content-Type") or file_path.endswith(".json"):
                return cls.load_json(encoded_config)
            else:
                return cls.load_yaml(encoded_config)
        else:
            file_path = Path(file_path)
            encoded_config = file_path.read_text(encoding="utf8")

            if file_path.name.endswith(".json"):
                return cls.load_json(encoded_config)
            else:
                return cls.load_yaml(encoded_config)

    @classmethod
    def load_json(cls, json: str):
        """Instantiate the config from a JSON string"""
        return cls.load_dict(config=jsonlib.loads(json))

    @classmethod
    def load_yaml(cls, yaml: str):
        """Instantiate the config from a YAML string"""
        config = yamllib.safe_load(yaml)
        if not isinstance(config, dict):
            raise UnexpectedConfigurationFormat(
                f"The config file was loaded but it appears to be in an unexpected format. "
                f"The root of the configuration should be a key/value mapping, but the "
                f"type '{type(config).__name__}' was found instead. Check your config "
                f"file is correctly formatted."
            )
        return cls.load_dict(config=config)

    @classmethod
    def load_dict(cls, config: dict, set_defaults=True):
        """Instantiate the config from a dictionary"""
        # pylint: disable=import-outside-toplevel
        from .structure import RootConfig

        config = config.copy()
        if set_defaults:
            config = set_default_config(config)
        validate_config(config)

        return cls(root_config=cast_to_hint(config, RootConfig))

    @classmethod
    def default(cls):
        return cls.load_dict(config={}, set_defaults=True)

    def __getattr__(self, item):
        if hasattr(self._config, item):
            return getattr(self._config, item)
        else:
            raise AttributeError(f"No root-level configuration option named '{item}'")


def validate_config(config: dict):
    """Validate the provided config dictionary against the config json schema"""
    json_schema = config_as_json_schema()
    jsonschema.validate(config, json_schema)


def config_as_json_schema() -> dict:
    """Get the configuration structure as a json schema"""
    # pylint: disable=import-outside-toplevel
    from .structure import RootConfig

    schema, = python_type_to_json_schemas(RootConfig)
    # Some of the default values will still be python types,
    # so let's use deform_to_bus to turn them into something
    # that'll be json safe
    schema = deform_to_bus(schema)

    schema["$schema"] = SCHEMA_URI
    return schema


def set_default_config(config: dict) -> dict:
    """Set the default configuration options on a loaded config dictionary"""
    env_service_name = os.environ.get("LIGHTBUS_SERVICE_NAME")
    if env_service_name:
        config.setdefault("service_name", env_service_name)

    env_process_name = os.environ.get("LIGHTBUS_PROCESS_NAME")
    if env_process_name:
        config.setdefault("process_name", env_process_name)

    config.setdefault("apis", {})
    config.setdefault("bus", {})
    config["apis"].setdefault("default", {})
    config["bus"].setdefault("schema", {})

    config["apis"]["default"].setdefault("rpc_transport", {"redis": {}})
    config["apis"]["default"].setdefault("result_transport", {"redis": {}})
    config["apis"]["default"].setdefault("event_transport", {"redis": {}})
    config["bus"]["schema"].setdefault("transport", {"redis": {}})
    return config



from lightbus.api import Api, Event


class LightbusStateApi(Api):
    """The API for the state plugin"""

    worker_started = Event(
        parameters=[
            "service_name",
            "process_name",
            "metrics_enabled",
            "api_names",
            "listening_for",
            "timestamp",
            "ping_interval",
        ]
    )
    worker_ping = Event(
        parameters=[
            "service_name",
            "process_name",
            "metrics_enabled",
            "api_names",
            "listening_for",
            "timestamp",
            "ping_interval",
        ]
    )
    worker_stopped = Event(parameters=["process_name", "timestamp"])

    class Meta:
        name = "internal.state"
        internal = True


class LightbusMetricsApi(Api):
    """The API for the metrics plugin"""

    rpc_call_sent = Event(
        parameters=[
            "service_name",
            "process_name",
            "id",
            "api_name",
            "procedure_name",
            "kwargs",
            "timestamp",
        ]
    )
    rpc_call_received = Event(
        parameters=["service_name", "process_name", "id", "api_name", "procedure_name", "timestamp"]
    )
    rpc_response_sent = Event(
        parameters=[
            "service_name",
            "process_name",
            "id",
            "api_name",
            "procedure_name",
            "result",
            "timestamp",
        ]
    )
    rpc_response_received = Event(
        parameters=["service_name", "process_name", "id", "api_name", "procedure_name", "timestamp"]
    )

    event_fired = Event(
        parameters=[
            "service_name",
            "process_name",
            "event_id",
            "api_name",
            "event_name",
            "kwargs",
            "timestamp",
        ]
    )
    event_received = Event(
        parameters=[
            "service_name",
            "process_name",
            "event_id",
            "api_name",
            "event_name",
            "kwargs",
            "timestamp",
        ]
    )
    event_processed = Event(
        parameters=[
            "service_name",
            "process_name",
            "event_id",
            "api_name",
            "event_name",
            "kwargs",
            "timestamp",
        ]
    )

    class Meta:
        name = "internal.metrics"
        internal = True



import sys

if sys.version_info >= (3, 8):
    from functools import singledispatchmethod
else:
    from functools import singledispatch, update_wrapper

    def singledispatchmethod(func):
        dispatcher = singledispatch(func)

        def wrapper(*args, **kw):
            return dispatcher.dispatch(args[1].__class__)(*args, **kw)

        wrapper.register = dispatcher.register
        update_wrapper(wrapper, func)
        return wrapper



from lightbus.transports.base import (
    RpcTransport,
    ResultTransport,
    EventTransport,
    SchemaTransport,
    Transport,
)
from lightbus.transports.debug import (
    DebugRpcTransport,
    DebugResultTransport,
    DebugEventTransport,
    DebugSchemaTransport,
)
from lightbus.transports.redis.rpc import RedisRpcTransport
from lightbus.transports.redis.result import RedisResultTransport
from lightbus.transports.redis.event import RedisEventTransport
from lightbus.transports.redis.schema import RedisSchemaTransport



import argparse
import asyncio
import logging
import os
import signal
import sys

from lightbus import BusPath
from lightbus.commands import utilities as command_utilities
from lightbus.plugins import PluginRegistry
from lightbus.utilities.async_tools import block
from lightbus.utilities.features import Feature, ALL_FEATURES
from lightbus.utilities.logging import log_welcome_message

logger = logging.getLogger(__name__)

csv_type = lambda value: [v.strip() for v in value.split(",") if value.split(",")]


class Command:
    def setup(self, parser, subparsers):
        # pylint: disable=attribute-defined-outside-init
        self.all_features = [f.value for f in Feature]
        self.features_str = ", ".join(self.all_features)

        parser_run = subparsers.add_parser(
            "run", help="Run Lightbus", formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )

        group = parser_run.add_argument_group(title="Run command arguments")
        group.add_argument(
            "--only",
            "-o",
            help=f"Only provide the specified features. Comma separated list. Possible values: {self.features_str}",
            type=csv_type,
        )
        group.add_argument(
            "--skip",
            "-k",
            help=(
                f"Provide all except the specified features. Comma separated list. "
                f"Possible values: {self.features_str}"
            ),
            type=csv_type,
        )
        group.add_argument(
            "--schema",
            "-m",
            help=(
                "Manually load the schema from the given file or directory. "
                "This will normally be provided by the schema transport, "
                "but manual loading may be useful during development or testing."
            ),
            metavar="FILE_OR_DIRECTORY",
        )
        parser_run.set_defaults(func=self.handle)

        command_utilities.setup_common_arguments(parser_run)

    def handle(self, args, config, plugin_registry: PluginRegistry):
        try:
            self._handle(args, config, plugin_registry)
        except Exception as e:
            block(plugin_registry.execute_hook("exception", e=e), timeout=5)
            raise

    def _handle(self, args, config, plugin_registry: PluginRegistry):
        command_utilities.setup_logging(override=getattr(args, "log_level", None), config=config)

        bus: BusPath
        bus_module, bus = command_utilities.import_bus(args)

        # Convert only & skip into a list of features to enable
        if args.only or args.skip:
            if args.only:
                features = args.only
            else:
                features = self.all_features

            for skip_feature in args.skip or []:
                if skip_feature in features:
                    features.remove(skip_feature)
        elif os.environ.get("LIGHTBUS_FEATURES"):
            features = csv_type(os.environ.get("LIGHTBUS_FEATURES"))
        else:
            features = ALL_FEATURES

        bus.client.set_features(features)

        # TODO: Move to lightbus.create()?
        if args.schema:
            if args.schema == "-":
                # if '-' read from stdin
                source = None
            else:
                source = args.schema
            bus.schema.load_local(source)

        restart_signals = (signal.SIGINT, signal.SIGTERM)

        # Handle incoming signals
        async def signal_handler():
            # Stop handling signals now. If we receive the signal again
            # let the process quit naturally
            for signal_ in restart_signals:
                asyncio.get_event_loop().remove_signal_handler(signal_)

            logger.debug("Caught signal. Stopping Lightbus worker")
            bus.client.request_shutdown()

        for signal_ in restart_signals:
            asyncio.get_event_loop().add_signal_handler(
                signal_, lambda: asyncio.ensure_future(signal_handler())
            )

        log_welcome_message(
            logger=logger,
            transport_registry=bus.client.transport_registry,
            schema=bus.client.schema,
            plugin_registry=bus.client.plugin_registry,
            config=bus.client.config,
        )

        try:
            block(plugin_registry.execute_hook("receive_args", args=args), timeout=5)
            bus.client.run_forever()

        finally:
            # Cleanup signal handlers
            for signal_ in restart_signals:
                asyncio.get_event_loop().remove_signal_handler(signal_)

        if bus.client.exit_code:
            sys.exit(bus.client.exit_code)



import argparse
import logging

from lightbus.commands import utilities as command_utilities
from lightbus.config.config import config_as_json_schema
from lightbus.plugins import PluginRegistry
from lightbus.schema.encoder import json_encode

logger = logging.getLogger(__name__)


class Command:
    def setup(self, parser, subparsers):
        parser_dumpconfigschema = subparsers.add_parser(
            "dumpconfigschema",
            help=(
                "Dumps the lightbus configuration json schema. Can be useful "
                "in validating your config. This is not the same as your "
                "bus' API schema, for that see the more commonly used 'dumpschema' "
                "command"
            ),
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        group = parser_dumpconfigschema.add_argument_group(
            title="Dump config schema command arguments"
        )
        group.add_argument(
            "--out",
            "-o",
            help=(
                "File to write config schema to. "
                "If omitted the schema will be written to standard out."
            ),
            metavar="FILE",
        )
        parser_dumpconfigschema.set_defaults(func=self.handle)

    def handle(self, args):
        schema = json_encode(config_as_json_schema(), indent=2, sort_keys=True)

        if args.out:
            with open(args.out, "w", encoding="utf8") as f:
                f.write(schema)
        else:
            print(schema)



import asyncio
import logging
from functools import wraps
from inspect import iscoroutinefunction

from typing import List

from lightbus import RpcTransport, Api, RpcMessage
from lightbus.client.docks.base import BaseDock
from lightbus.client.utilities import queue_exception_checker
from lightbus.client import commands
from lightbus.exceptions import TransportIsClosed
from lightbus.transports.pool import TransportPool
from lightbus.utilities.async_tools import cancel
from lightbus.utilities.internal_queue import InternalQueue
from lightbus.utilities.singledispatch import singledispatchmethod

logger = logging.getLogger(__name__)


class RpcResultDock(BaseDock):
    """ Takes internal Lightbus commands and performs interactions with the RPC & Result transport
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.consumer_tasks = set()
        self.listener_tasks = set()

    @singledispatchmethod
    async def handle(self, command):
        raise NotImplementedError(f"Did not recognise command {command.__class__.__name__}")

    @handle.register
    async def handle_consume_rpcs(self, command: commands.ConsumeRpcsCommand):
        """Worker wishes for incoming RPC results to be listened for and processed"""
        # Not all APIs will necessarily be served by the same transport, so group them
        # accordingly
        api_names_by_transport = self.transport_registry.get_rpc_transports(command.api_names)

        for rpc_transport, transport_api_names in api_names_by_transport.items():
            transport_apis = list(map(self.api_registry.get, transport_api_names))

            # fmt: off
            task = asyncio.ensure_future(queue_exception_checker(
                self._consume_rpcs_with_transport(rpc_transport=rpc_transport, apis=transport_apis),
                self.error_queue,
            ))
            # fmt: on
            self.consumer_tasks.add(task)

    @handle.register
    async def handle_call_rpc(self, command: commands.CallRpcCommand):
        """Client wishes to call a remote RPC"""
        api_name = command.message.api_name

        rpc_transport = self.transport_registry.get_rpc_transport(api_name)
        result_transport = self.transport_registry.get_result_transport(api_name)

        # TODO: Perhaps move return_path out of RpcMessage, as this feels a little gross
        command.message.return_path = await result_transport.get_return_path(command.message)

        await rpc_transport.call_rpc(command.message, options=command.options)

    @handle.register
    async def handle_send_result(self, command: commands.SendResultCommand):
        """Worker wishes to sent a result back to a client"""
        result_transport = self.transport_registry.get_result_transport(command.message.api_name)
        return await result_transport.send_result(
            command.rpc_message, command.message, command.rpc_message.return_path
        )

    @handle.register
    async def handle_receive_result(self, command: commands.ReceiveResultCommand):
        """Client wishes to receive a result from a worker"""

        # TODO: rpc_timeout is in three different places in the config!
        #       Fix this. Really it makes most sense for the use if it goes on the
        #       ApiConfig rather than having to repeat it on both the result & RPC
        #       transports.
        timeout = command.options.get(
            "timeout", self.config.api(command.message.api_name).rpc_timeout
        )
        result_transport = self.transport_registry.get_result_transport(command.message.api_name)

        logger.debug("Starting RPC result listener")
        task = asyncio.ensure_future(
            queue_exception_checker(
                self._result_listener(
                    result_transport=result_transport,
                    timeout=timeout,
                    rpc_message=command.message,
                    return_path=command.message.return_path,
                    options=command.options,
                    result_queue=command.destination_queue,
                ),
                self.error_queue,
            )
        )
        self.listener_tasks.add(task)

    async def _result_listener(
        self,
        result_transport: TransportPool,
        timeout: float,
        rpc_message: RpcMessage,
        return_path: str,
        options: dict,
        result_queue: InternalQueue,
    ):
        try:
            logger.debug("Result listener is waiting")
            result = await asyncio.wait_for(
                result_transport.receive_result(
                    rpc_message=rpc_message, return_path=return_path, options=options
                ),
                timeout=timeout,
            )
        except asyncio.TimeoutError as e:
            logger.debug("Result listener timed out")
            await result_queue.put(e)
        else:
            logger.debug("Result listener received result, putting onto result queue")
            await result_queue.put(result)
        finally:
            self.listener_tasks.discard(asyncio.current_task())

    @handle.register
    async def handle_close(self, command: commands.CloseCommand):
        """Client or worker wishes us to close down"""
        await cancel(*self.consumer_tasks)
        await cancel(*self.listener_tasks)

        for rpc_transport in self.transport_registry.get_all_rpc_transports():
            await rpc_transport.close()
        for rpc_transport in self.transport_registry.get_all_result_transports():
            await rpc_transport.close()

        await self.consumer.close()
        await self.producer.close()

    async def _consume_rpcs_with_transport(self, rpc_transport: RpcTransport, apis: List[Api]):
        while True:
            try:
                rpc_messages = await rpc_transport.consume_rpcs(apis)
            except TransportIsClosed:
                return

            for rpc_message in rpc_messages:
                await self.producer.send(commands.ExecuteRpcCommand(message=rpc_message)).wait()



import asyncio
from collections import defaultdict
from typing import Dict, NamedTuple, List, Callable, Optional

from lightbus.client.utilities import queue_exception_checker, ErrorQueueType
from lightbus.utilities.async_tools import run_user_provided_callable


class CallbackKey(NamedTuple):
    name: str
    run_before_plugins: bool


class HookRegistry:
    def __init__(
        self,
        error_queue: ErrorQueueType,
        execute_plugin_hooks: Callable,
        extra_parameters: Optional[dict] = None,
    ):
        self._hook_callbacks: Dict[CallbackKey, List[Callable]] = defaultdict(list)
        self.error_queue = error_queue
        self.execute_plugin_hooks = execute_plugin_hooks
        self.extra_parameters = extra_parameters or {}

    def set_extra_parameter(self, name, value):
        self.extra_parameters[name] = value

    async def execute(self, name, **kwargs):
        # Hooks that need to run before plugins
        key = CallbackKey(name, run_before_plugins=True)
        for callback in self._hook_callbacks[key]:
            await queue_exception_checker(
                run_user_provided_callable(
                    callback, args=[], kwargs=dict(**self.extra_parameters, **kwargs)
                ),
                self.error_queue,
            )

        await self.execute_plugin_hooks(name, **self.extra_parameters, **kwargs)

        # Hooks that need to run after plugins
        key = CallbackKey(name, run_before_plugins=False)
        for callback in self._hook_callbacks[key]:
            await run_user_provided_callable(
                callback, args=[], kwargs=dict(**self.extra_parameters, **kwargs)
            )

    def register_callback(self, name, fn, before_plugins=False):
        key = CallbackKey(name, bool(before_plugins))
        self._hook_callbacks[key].append(fn)



"""Plugin to broadcast Lightbus' state on the internal.state API"""
from datetime import datetime
from typing import Coroutine, TYPE_CHECKING

from lightbus.message import EventMessage, RpcMessage, ResultMessage
from lightbus.plugins import LightbusPlugin
from lightbus.utilities.deforming import deform_to_bus

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    import lightbus
    from lightbus.config import Config


class MetricsPlugin(LightbusPlugin):
    priority = 110

    def __init__(self, service_name: str, process_name: str):
        self.service_name = service_name
        self.process_name = process_name

    @classmethod
    def from_config(cls, config: "Config"):
        return cls(service_name=config.service_name, process_name=config.process_name)

    # Client-side RPC hooks

    async def before_rpc_call(
        self, *, rpc_message: RpcMessage, client: "lightbus.client.BusClient"
    ):
        await self.send_event(
            client,
            "rpc_call_sent",
            id=rpc_message.id,
            api_name=rpc_message.api_name,
            procedure_name=rpc_message.procedure_name,
            kwargs=deform_to_bus(rpc_message.kwargs),
        )

    async def after_rpc_call(
        self,
        *,
        rpc_message: RpcMessage,
        result_message: ResultMessage,
        client: "lightbus.client.BusClient",
    ):
        await self.send_event(
            client,
            "rpc_response_received",
            id=rpc_message.id,
            api_name=rpc_message.api_name,
            procedure_name=rpc_message.procedure_name,
        )

    # Worker-side RPC hooks

    async def before_rpc_execution(
        self, *, rpc_message: RpcMessage, client: "lightbus.client.BusClient"
    ):
        await self.send_event(
            client,
            "rpc_call_received",
            id=rpc_message.id,
            api_name=rpc_message.api_name,
            procedure_name=rpc_message.procedure_name,
        )

    async def after_rpc_execution(
        self,
        *,
        rpc_message: RpcMessage,
        result_message: ResultMessage,
        client: "lightbus.client.BusClient",
    ):
        await self.send_event(
            client,
            "rpc_response_sent",
            id=rpc_message.id,
            api_name=rpc_message.api_name,
            procedure_name=rpc_message.procedure_name,
            result=deform_to_bus(result_message.result),
        )

    # Client-side event hooks

    async def after_event_sent(
        self, *, event_message: EventMessage, client: "lightbus.client.BusClient"
    ):
        if event_message.api_name == "internal.metrics":
            return

        await self.send_event(
            client,
            "event_fired",
            event_id="event_id",
            api_name=event_message.api_name,
            event_name=event_message.event_name,
            kwargs=deform_to_bus(event_message.kwargs),
        )

    # Worker-side event hooks

    async def before_event_execution(
        self, *, event_message: EventMessage, client: "lightbus.client.BusClient"
    ):
        if event_message.api_name == "internal.metrics":
            return

        await self.send_event(
            client,
            "event_received",
            event_id="event_id",
            api_name=event_message.api_name,
            event_name=event_message.event_name,
            kwargs=deform_to_bus(event_message.kwargs),
        )

    async def after_event_execution(
        self, *, event_message: EventMessage, client: "lightbus.client.BusClient"
    ):
        if event_message.api_name == "internal.metrics":
            return

        await self.send_event(
            client,
            "event_processed",
            event_id="event_id",
            api_name=event_message.api_name,
            event_name=event_message.event_name,
            kwargs=deform_to_bus(event_message.kwargs),
        )

    def send_event(self, client: "lightbus.client.BusClient", event_name_, **kwargs) -> Coroutine:
        """Send an event to the bus
        """
        kwargs.setdefault("timestamp", datetime.utcnow().timestamp())
        kwargs.setdefault("service_name", self.service_name)
        kwargs.setdefault("process_name", self.process_name)
        kwargs = deform_to_bus(kwargs)

        return client.fire_event(
            api_name="internal.metrics", name=event_name_, kwargs=kwargs, options={}
        )



import asyncio
import logging
from datetime import datetime, timezone
from typing import Optional, Mapping

import aioredis
from aioredis import Redis, ConnectionsPool, ReplyError, ConnectionClosedError

from aioredis.util import decode

from lightbus.transports.base import EventMessage
from lightbus.exceptions import LightbusException, TransportIsClosed, LightbusShutdownInProgress
from lightbus.utilities.frozendict import frozendict

logger = logging.getLogger("lightbus.transports.redis")


def redis_stream_id_subtract_one(message_id):
    """Subtract one from the message ID

    This is useful when we need to xread() events inclusive of the given ID,
    rather than exclusive of the given ID (which is the sensible default).
    Only use when one can tolerate an exceptionally slim risk of grabbing extra events.
    """
    milliseconds, n = map(int, message_id.split("-"))
    if n > 0:
        n = n - 1
    elif milliseconds > 0:
        milliseconds = milliseconds - 1
        n = 9999
    else:
        # message_id is '0000000000000-0'. Subtracting one
        # from this is neither possible, desirable or useful.
        return message_id
    return "{:13d}-{}".format(milliseconds, n)


def redis_stream_id_add_one(message_id):
    """Add one to the message ID

    This is useful when we need to xrange() events exclusive of the given ID,
    rather than inclusive of the given ID (which is the sensible default).
    There is no chance of missing events with this method.
    """
    milliseconds, n = map(int, message_id.split("-"))
    return f"{milliseconds:013d}-{n + 1}"


def normalise_since_value(since):
    """Take a 'since' value and normalise it to be a redis message ID"""
    if not since:
        return "$"
    elif hasattr(since, "timestamp"):  # datetime
        # Create message ID: "<milliseconds-timestamp>-<sequence-number>"
        return f"{round(since.timestamp() * 1000):013d}-0"
    else:
        return since


def redis_steam_id_to_datetime(message_id):
    message_id = decode(message_id, "utf8")
    milliseconds, seq = map(int, message_id.split("-"))
    # Treat the sequence value as additional microseconds to ensure correct sequencing
    microseconds = (milliseconds % 1000 * 1000) + seq
    dt = datetime.utcfromtimestamp(milliseconds // 1000).replace(
        microsecond=microseconds, tzinfo=timezone.utc
    )
    return dt


def datetime_to_redis_steam_id(dt: datetime) -> str:
    timestamp = round(dt.timestamp() * 1000)
    # We ignore microseconds here as using them in the sequence
    # number would make the the sequence number non sequential
    return f"{timestamp:013d}-0"


class InvalidRedisPool(LightbusException):
    pass


class RedisEventMessage(EventMessage):
    def __init__(self, *, stream: str, native_id: str, consumer_group: Optional[str], **kwargs):
        super(RedisEventMessage, self).__init__(**kwargs)
        self.stream = stream
        self.native_id = native_id
        self.consumer_group = consumer_group

    @property
    def datetime(self):
        return redis_steam_id_to_datetime(self.native_id)

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()
        metadata["stream"] = self.stream
        metadata["native_id"] = self.native_id
        metadata["consumer_group"] = self.consumer_group
        return metadata


class RedisTransportMixin:
    connection_parameters: dict = {"address": "redis://localhost:6379", "maxsize": 100}
    _redis_pool = None

    def set_redis_pool(
        self,
        redis_pool: Optional[Redis],
        url: str = None,
        connection_parameters: Mapping = frozendict(),
    ):
        self._redis_pool = None
        self._closed = False

        if not redis_pool:
            # Connect lazily using the provided parameters

            self.connection_parameters = self.connection_parameters.copy()
            self.connection_parameters.update(connection_parameters)
            if url:
                self.connection_parameters["address"] = url
        else:
            # Use the provided connection

            if isinstance(redis_pool, (ConnectionsPool,)):
                # If they've passed a raw pool then wrap it up in a Redis object.
                # aioredis.create_redis_pool() normally does this for us.
                redis_pool = Redis(redis_pool)
            if not isinstance(redis_pool, (Redis,)):
                raise InvalidRedisPool(
                    "Invalid Redis connection provided: {}. If unsure, use aioredis.create_redis_pool() to "
                    "create your redis connection.".format(redis_pool)
                )
            if not isinstance(redis_pool._pool_or_conn, (ConnectionsPool,)):
                raise InvalidRedisPool(
                    "The provided redis connection is backed by a single connection, rather than a "
                    "pool of connections. This will lead to lightbus deadlocks and is unsupported. "
                    "If unsure, use aioredis.create_redis_pool() to create your redis connection."
                )

            # Determine the connection parameters from the given pool
            # (we will need these in other to create new pools for other threads)
            self.connection_parameters = dict(
                address=redis_pool.address,
                db=redis_pool.db,
                password=redis_pool._pool_or_conn._password,
                encoding=redis_pool.encoding,
                minsize=redis_pool._pool_or_conn.minsize,
                maxsize=redis_pool._pool_or_conn.maxsize,
                ssl=redis_pool._pool_or_conn._ssl,
                parser=redis_pool._pool_or_conn._parser_class,
                timeout=redis_pool._pool_or_conn._create_connection_timeout,
                connection_cls=redis_pool._pool_or_conn._connection_cls,
            )

            self._redis_pool = redis_pool

    async def connection_manager(self) -> Redis:
        if self._closed:
            # This was first caught when the state plugin tried to send a
            # message to the bus on upon the after_worker_stopped stopped event.
            raise TransportIsClosed(
                "Transport has been closed. Connection to Redis is no longer available."
            )

        if not self._redis_pool:
            self._redis_pool = await aioredis.create_redis_pool(**self.connection_parameters)

        try:
            internal_pool = self._redis_pool._pool_or_conn
            if hasattr(internal_pool, "size") and hasattr(internal_pool, "maxsize"):
                if internal_pool.size == internal_pool.maxsize:
                    logging.critical(
                        "Redis pool has reached maximum size. It is possible that this will recover normally, "
                        "you may have more event listeners than connections available to the Redis pool. "
                        "You can increase the redis pool size by specifying the `maxsize` "
                        "parameter in each of the Redis transport configuration sections. Current maxsize is: {}"
                        "".format(self.connection_parameters.get("maxsize"))
                    )

            return await self._redis_pool
        except aioredis.PoolClosedError:
            raise LightbusShutdownInProgress(
                "Redis connection pool has been closed. Assuming shutdown in progress."
            )

    async def close(self):
        self._closed = True
        if self._redis_pool:
            await self._close_redis_pool()

    async def _close_redis_pool(self):
        self._redis_pool.close()
        await self._redis_pool.wait_closed()
        # In Python >= 3.7 the redis connections do not actually get
        # immediately closed following the above call to wait_closed().
        # If you need to be sure the the connections have closed the
        # a call to `await asyncio.sleep(0.001)` does the trick
        del self._redis_pool

    def __str__(self):
        if self._redis_pool:
            conn = self._redis_pool.connection
            return f"redis://{conn.address[0]}:{conn.address[1]}/{conn.db}"
        else:
            return self.connection_parameters.get("address", "Unknown URL")


async def retry_on_redis_connection_failure(fn, *, args=tuple(), retry_delay: int, action: str):
    """Decorator to repeatedly call fn in case of Redis connection problems"""
    while True:
        try:
            return await fn(*args)
        except (ConnectionClosedError, ConnectionResetError):
            # ConnectionClosedError is from aioredis. However, sometimes the connection
            # can die outside of aioredis, in which case we get a builtin ConnectionResetError.
            logger.warning(
                f"Redis connection lost while {action}, reconnecting "
                f"in {retry_delay} seconds..."
            )
            await asyncio.sleep(retry_delay)
        except ConnectionRefusedError:
            logger.warning(
                f"Redis connection refused while {action}, retrying " f"in {retry_delay} seconds..."
            )
            await asyncio.sleep(retry_delay)
        except ReplyError as e:
            if "LOADING" in str(e):
                logger.warning(
                    f"Redis server is still loading, retrying " f"in {retry_delay} seconds..."
                )
                await asyncio.sleep(retry_delay)
            else:
                raise



""" Serializers suitable for transports which require a single serialised value

These serializers handle moving data to/from a string-based format.

"""
from typing import Union, TYPE_CHECKING

from lightbus.serializers.base import (
    decode_bytes,
    sanity_check_metadata,
    MessageSerializer,
    MessageDeserializer,
)

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from lightbus import Message


class BlobMessageSerializer(MessageSerializer):
    def __call__(self, message: "Message") -> str:
        """Takes a message object and returns a serialised string representation"""
        return self.encoder({"metadata": message.get_metadata(), "kwargs": message.get_kwargs()})


class BlobMessageDeserializer(MessageDeserializer):
    def __call__(self, serialized: Union[str, dict], *, native_id=None, **extra):
        """ Takes a serialised string representation and returns a Message object

        Reverse of BlobMessageSerializer
        """
        # Allow for receiving dicts on the assumption that this will be
        # json which has already been decoded.
        if isinstance(serialized, dict):
            decoded = serialized
        else:
            serialized = decode_bytes(serialized)
            decoded = self.decoder(serialized)

        metadata = decoded.get("metadata", {})
        kwargs = decoded.get("kwargs", {})

        sanity_check_metadata(self.message_class, metadata)

        return self.message_class.from_dict(
            metadata=metadata, kwargs=kwargs, native_id=native_id, **extra
        )



import inspect
from typing import Optional, Type, Union, Tuple, List, TypeVar


def type_is_namedtuple(t) -> bool:
    """Figuring out if a type is a named tuple is not as trivial as one may expect"""
    try:
        return issubclass(t, tuple) and hasattr(t, "_fields")
    except TypeError:
        return False


def is_namedtuple(v) -> bool:
    """Figuring out if an object is a named tuple is not as trivial as one may expect"""
    try:
        return isinstance(v, tuple) and hasattr(v, "_fields")
    except TypeError:
        return False


def type_is_dataclass(t) -> bool:
    return hasattr(t, "__dataclass_fields__")


def is_dataclass(v) -> bool:
    return hasattr(v, "__dataclass_fields__")


def is_optional(hint) -> Optional[Type]:
    hint_type, hint_args = parse_hint(hint)
    if hint_type == Union and len(hint_args) == 2 and hint_args[1] == type(None):
        return hint_args[0]
    else:
        return None


def isinstance_safe(value, type_):
    """Determine if value is a subclass of type_

    Will work even if it is not a valid question to ask of
    the given value.
    """
    try:
        return isinstance(value, type_)
    except TypeError:
        # Cannot perform isinstance on some types
        return False


def issubclass_safe(value, type_):
    """Determine if value is a subclass of type_

    Will work even value is not a class
    """
    try:
        return issubclass(value, type_)
    except (TypeError, AttributeError):
        # Cannot perform issubclass on some types
        return False


def parse_hint(hint: Type) -> Tuple[Type, Optional[List]]:
    """ Parse a typing hint into its type and and arguments.

    For example:

        >>> parse_hint(Union[dict, list])
        (typing.Union, [<class 'dict'>, <class 'list'>])

        >>> parse_hint(int)
        (<class 'int'>, None)

    """
    if hasattr(hint, "__origin__"):
        # This is a type hint (eg typing.Union)
        # Filter out TypeVars such as KT & VT_co (they generally
        # indicate that no explicit hint was given)
        hint_args = [a for a in hint.__args__ if not isinstance(a, TypeVar)]
        return hint.__origin__, hint_args or None
    else:
        # This is something other than a type hint
        # (e.g. an int or datetime)
        return hint, None


def get_property_default(type_: Type, property_name: str) -> ...:
    """ Get the default value for a property with name `property_name` on class `type_`"""
    if issubclass_safe(type_, tuple):
        # namedtuple
        if hasattr(type_, "_field_defaults"):
            default = type_._field_defaults.get(property_name, inspect.Parameter.empty)
        else:
            default = inspect.Parameter.empty
    else:
        # everything else
        default = getattr(type_, property_name, inspect.Parameter.empty)

    if callable(default):
        default = inspect.Parameter.empty

    return default



"""Generation of human-friendly strings"""

import logging
import secrets

logger = logging.getLogger(__name__)

_adjectives = [
    "aged",
    "ancient",
    "autumn",
    "billowing",
    "bitter",
    "black",
    "blue",
    "bold",
    "broad",
    "broken",
    "calm",
    "cold",
    "cool",
    "crimson",
    "curly",
    "damp",
    "dark",
    "dawn",
    "delicate",
    "divine",
    "dry",
    "empty",
    "falling",
    "fancy",
    "flat",
    "floral",
    "fragrant",
    "frosty",
    "gentle",
    "green",
    "hidden",
    "holy",
    "icy",
    "jolly",
    "late",
    "lingering",
    "little",
    "lively",
    "long",
    "lucky",
    "misty",
    "morning",
    "muddy",
    "mute",
    "nameless",
    "noisy",
    "odd",
    "old",
    "orange",
    "patient",
    "plain",
    "polished",
    "proud",
    "purple",
    "quiet",
    "rapid",
    "raspy",
    "red",
    "restless",
    "rough",
    "round",
    "royal",
    "shiny",
    "shrill",
    "shy",
    "silent",
    "small",
    "snowy",
    "soft",
    "solitary",
    "sparkling",
    "spring",
    "square",
    "steep",
    "still",
    "summer",
    "super",
    "sweet",
    "throbbing",
    "tight",
    "tiny",
    "twilight",
    "wandering",
    "weathered",
    "white",
    "wild",
    "winter",
    "wispy",
    "withered",
    "yellow",
    "young",
]
_nouns = [
    "art",
    "band",
    "bar",
    "base",
    "bird",
    "block",
    "boat",
    "bonus",
    "bread",
    "breeze",
    "brook",
    "bush",
    "butterfly",
    "cake",
    "cell",
    "cherry",
    "cloud",
    "credit",
    "darkness",
    "dawn",
    "dew",
    "disk",
    "dream",
    "dust",
    "feather",
    "field",
    "fire",
    "firefly",
    "flower",
    "fog",
    "forest",
    "frog",
    "frost",
    "glade",
    "glitter",
    "grass",
    "hall",
    "hat",
    "haze",
    "heart",
    "hill",
    "king",
    "lab",
    "lake",
    "leaf",
    "limit",
    "math",
    "meadow",
    "mode",
    "moon",
    "morning",
    "mountain",
    "mouse",
    "mud",
    "night",
    "paper",
    "pine",
    "poetry",
    "pond",
    "queen",
    "rain",
    "recipe",
    "resonance",
    "rice",
    "river",
    "salad",
    "scene",
    "sea",
    "shadow",
    "shape",
    "silence",
    "sky",
    "smoke",
    "snow",
    "snowflake",
    "sound",
    "star",
    "sun",
    "sun",
    "sunset",
    "surf",
    "term",
    "thunder",
    "tooth",
    "tree",
    "truth",
    "union",
    "unit",
    "violet",
    "voice",
    "water",
    "waterfall",
    "wave",
    "wildflower",
    "wind",
    "wood",
]


def human_time(seconds: float):
    if seconds > 1:
        return "{} seconds".format(round(seconds, 2))
    else:
        return "{} milliseconds".format(round(seconds * 1000, 2))


def generate_human_friendly_name():
    return "{}-{}-{}".format(
        secrets.choice(_adjectives), secrets.choice(_nouns), secrets.randbelow(999) + 1
    )



from .utilities.logging import configure_logging
from .transports import *
from .client import *
from .path import *
from .message import *
from .api import *
from .schema import *
from .creation import *
from .client.utilities import OnError
from .exceptions import *



import inspect
import json
from typing import Union, TypeVar, Type

from lightbus.exceptions import InvalidMessage, InvalidSerializerConfiguration
from lightbus.schema.encoder import json_encode


def decode_bytes(b: Union[str, bytes]):
    return b.decode("utf8") if isinstance(b, bytes) else b


def sanity_check_metadata(message_class, metadata):
    """Takes unserialized metadata and checks it looks sane

    This relies upon the required_metadata of each Message class
    to provide a list of metadata fields that are required.
    """
    for required_key in message_class.required_metadata:
        if required_key not in metadata:
            raise InvalidMessage(
                "Required key '{key}' missing in {cls} metadata. "
                "Found keys: {keys}".format(
                    key=required_key, keys=", ".join(metadata.keys()), cls=message_class.__name__
                )
            )
        elif not metadata.get(required_key):
            raise InvalidMessage(
                "Required key '{key}' present in {cls} metadata but value was empty"
                "".format(key=required_key, cls=message_class.__name__)
            )


SerialisedData = TypeVar("SerialisedData")


class MessageSerializer:
    def __init__(self, encoder=json_encode):
        self.encoder = encoder

    def __call__(self, message: "lightbus.Message") -> SerialisedData:
        raise NotImplementedError()


class MessageDeserializer:
    def __init__(self, message_class: Type["lightbus.Message"], decoder=json.loads):
        if not inspect.isclass(message_class):
            raise InvalidSerializerConfiguration(
                "The message_class value provided to JsonMessageDeserializer was not a class, "
                "it was actually: {}".format(message_class)
            )

        self.message_class = message_class
        self.decoder = decoder

    def __call__(self, serialized: SerialisedData, *, native_id=None) -> "lightbus.Message":
        raise NotImplementedError()



import asyncio
import logging
from contextlib import ContextDecorator
from functools import wraps
from unittest.mock import patch, _patch

from typing import List, Dict, Tuple, TypeVar, Type, NamedTuple, Optional, Set

from lightbus import (
    RpcTransport,
    EventTransport,
    SchemaTransport,
    ResultTransport,
    RpcMessage,
    ResultMessage,
    EventMessage,
)
from lightbus.client.commands import SendEventCommand, CallRpcCommand
from lightbus.config import Config
from lightbus.path import BusPath
from lightbus.client import BusClient
from lightbus.transports.registry import TransportRegistry
from lightbus.utilities.internal_queue import InternalQueue

_registry: Dict[str, List] = {}
logger = logging.getLogger(__name__)


class MockResult:
    """Utility for mocking bus calls

    This is the context variable provided when using the BusMocker utility.

    Examples:

        # bus_mock will be a MockResult
        with bus_mocker as bus_mock:
            # ... do stuff with the bus ...
            bus_mock.assertEventFired()

    """

    def __init__(
        self, mocker_context: "BusQueueMockerContext", mock_responses: dict, mock_events: set
    ):
        self.mocker_context = mocker_context
        self.mock_responses = mock_responses
        self.mock_events = mock_events

    def assertEventFired(self, full_event_name, *, times=None):
        event_names_fired = self.eventNamesFired
        if times is None or times > 0:
            assert (  # nosec
                full_event_name in event_names_fired
            ), f"Event {full_event_name} was never fired. Fired events were: {', '.join(event_names_fired)}"

        if times is not None:
            total_times_fired = len([v for v in event_names_fired if v == full_event_name])
            assert total_times_fired == times, (  # nosec
                f"Event fired the incorrect number of times. "
                f"Expected {times}, actual {total_times_fired}"
            )

    assert_events_fired = assertEventFired

    def assertEventNotFired(self, full_event_name):
        assert (
            full_event_name not in self.eventNamesFired
        ), f"Event {full_event_name} was unexpectedly fired"

    assert_event_not_fired = assertEventNotFired

    def getEventMessages(self, full_event_name=None) -> List[EventMessage]:
        commands = self.mocker_context.event.to_transport.commands.get_all(SendEventCommand)
        if full_event_name is None:
            return [c.message for c in commands]
        else:
            return [c.message for c in commands if c.message.canonical_name == full_event_name]

    get_event_messages = getEventMessages

    def mockEventFiring(self, full_event_name: str):
        self.mock_events.add(full_event_name)

    mock_event_firing = mockEventFiring

    def assertRpcCalled(self, full_rpc_name, *, times=None):
        rpc_names_called = self.rpcNamesCalled
        if times is None or times > 0:
            assert (
                full_rpc_name in rpc_names_called
            ), f"RPC {full_rpc_name} was never called. Called RPCs were: {set(rpc_names_called)}"

        if times is not None:
            total_times_called = len([v for v in rpc_names_called if v == full_rpc_name])
            assert total_times_called == times, (  # nosec
                f"RPC {full_rpc_name} called the incorrect number of times. "
                f"Expected {times}, actual {total_times_called}"
            )

    assert_rpc_called = assertRpcCalled

    def assertRpcNotCalled(self, full_rpc_name):
        assert (
            full_rpc_name not in self.rpcNamesCalled
        ), f"Event {full_rpc_name} was unexpectedly fired"

    assert_rpc_not_called = assertRpcNotCalled

    def getRpcMessages(self, full_rpc_name=None) -> List[RpcMessage]:
        commands = self.mocker_context.rpc_result.to_transport.commands.get_all(CallRpcCommand)
        if full_rpc_name is None:
            return [c.message for c in commands]
        else:
            return [c.message for c in commands if c.message.canonical_name == full_rpc_name]

    get_rpc_messages = getRpcMessages

    def mockRpcCall(self, full_rpc_name, result=None, **rpc_result_message_kwargs):
        self.mock_responses[full_rpc_name] = dict(result=result, **rpc_result_message_kwargs)

    mock_rpc_call = mockRpcCall

    @property
    def eventNamesFired(self) -> List[str]:
        return [c.canonical_name for c in self.getEventMessages()]

    event_names_fired = eventNamesFired

    @property
    def rpcNamesCalled(self) -> List[str]:
        return [c.canonical_name for c in self.getRpcMessages()]

    rpc_names_called = rpcNamesCalled

    def __repr__(self):
        return f"<MockResult: events: {len(self.getEventMessages())}, rpcs: {len(self.getRpcMessages())}>"


class BusMocker(ContextDecorator):
    def __init__(self, bus: BusPath, require_mocking=True):
        self.bus = bus
        self.old_transport_registry = None
        self.require_mocking = require_mocking
        self.stack: List[BusQueueMockerContext] = []

    def __call__(self, func):
        # Overriding ContextDecorator.__call__ to pass the mock
        # result to the function. This is exactly the same as the parent
        # implementation except we pass mock_result into func().
        @wraps(func)
        def inner(*args, **kwds):
            with self._recreate_cm() as mock_result:
                return func(*args, mock_result, **kwds)

        return inner

    def __enter__(self):
        """Start of a context where all the bus' transports have been replaced with mocks"""

        # Mutable structures which we will use this to get the mocked data into
        # the transports

        # RPC
        mock_responses = {}
        # Events
        mock_events = set()

        # Create our transport classes, into which we inject our mutable structures
        TestRpcTransport = make_test_rpc_transport()
        TestResultTransport = make_test_result_transport(mock_responses)
        TestEventTransport = make_test_event_transport(mock_events)
        TestSchemaTransport = make_test_schema_transport()

        new_registry = TransportRegistry()
        new_registry.set_schema_transport(
            TestSchemaTransport, TestSchemaTransport.Config(), self.bus.client.config
        )

        self.old_transport_registry = self.bus.client.transport_registry

        for api_name, entry in self.old_transport_registry._registry.items():
            new_registry.set_rpc_transport(
                api_name, TestRpcTransport, TestRpcTransport.Config(), self.bus.client.config
            )
            new_registry.set_result_transport(
                api_name,
                TestResultTransport,
                TestResultTransport.Config(require_mocking=self.require_mocking),
                self.bus.client.config,
            )
            new_registry.set_event_transport(
                api_name,
                TestEventTransport,
                TestEventTransport.Config(require_mocking=self.require_mocking),
                self.bus.client.config,
            )

        # The docs are only available on the bus client during testing
        self.bus.client.event_dock.transport_registry = new_registry
        self.bus.client.rpc_result_dock.transport_registry = new_registry

        queue_mocker = BusQueueMockerContext(client=self.bus.client)
        bus_with_mocked_queues = queue_mocker.__enter__()
        self.stack.append(queue_mocker)
        return MockResult(
            bus_with_mocked_queues, mock_responses=mock_responses, mock_events=mock_events
        )

    def __exit__(self, exc_type, exc, exc_tb):
        """Restores the bus back to its original state"""
        bus_with_mocked_queues = self.stack.pop()
        bus_with_mocked_queues.__exit__(exc_type, exc, exc_tb)

        # The docs are only available on the bus client during testing
        bus_with_mocked_queues.client.event_dock.transport_registry = self.old_transport_registry
        bus_with_mocked_queues.client.rpc_result_dock.transport_registry = (
            self.old_transport_registry
        )


bus_mocker = BusMocker

T = TypeVar("T")


def make_test_rpc_transport():
    class TestRpcTransport(RpcTransport):
        def __init__(self):
            self.rpcs: List[Tuple[RpcMessage, dict]] = []

        async def call_rpc(self, rpc_message, options: dict):
            pass

        async def consume_rpcs(self, apis):
            raise NotImplementedError("Not yet supported by mocks")

    return TestRpcTransport


def make_test_result_transport(mock_responses: Dict[str, dict]):
    class TestResultTransport(ResultTransport):
        _mock_responses: Dict[str, dict] = mock_responses

        def __init__(self, require_mocking=True):
            super().__init__()
            self.mock_responses = mock_responses
            self.require_mocking = require_mocking

        @classmethod
        def from_config(cls: Type[T], config: "Config", require_mocking: bool = True) -> T:
            return cls(require_mocking=require_mocking)

        async def get_return_path(self, rpc_message):
            return "test://"

        async def send_result(self, rpc_message, result_message, return_path):
            raise NotImplementedError("Not yet supported by mocks")

        async def receive_result(self, rpc_message, return_path, options):
            if self.require_mocking:
                assert rpc_message.canonical_name in self.mock_responses, (
                    f"RPC {rpc_message.canonical_name} unexpectedly called. "
                    f"Perhaps you need to use mockRpcCall() to ensure the mocker expects this call."
                )

            if rpc_message.canonical_name in self.mock_responses:
                kwargs = self.mock_responses[rpc_message.canonical_name].copy()
                kwargs.setdefault("api_name", rpc_message.api_name)
                kwargs.setdefault("procedure_name", rpc_message.procedure_name)
                kwargs.setdefault("rpc_message_id", rpc_message.id)
                return ResultMessage(**kwargs)
            else:
                return ResultMessage(
                    result=None,
                    rpc_message_id="1",
                    api_name=rpc_message.api_name,
                    procedure_name=rpc_message.procedure_name,
                )

    return TestResultTransport


def make_test_event_transport(mock_events: set):
    class TestEventTransport(EventTransport):
        _mock_events: set = mock_events

        def __init__(self, require_mocking=True):
            super().__init__()
            self.events = []
            self.mock_events = mock_events
            self.require_mocking = require_mocking

        @classmethod
        def from_config(cls: Type[T], config: "Config", require_mocking: bool = True) -> T:
            return cls(require_mocking=require_mocking)

        async def send_event(self, event_message, options):
            if self.require_mocking:
                assert event_message.canonical_name in self.mock_events, (
                    f"Event {event_message.canonical_name} unexpectedly fired. "
                    f"Perhaps you need to use mockEventFiring() to ensure the mocker expects this call."
                )

        async def consume(self, listen_for: List[Tuple[str, str]], listener_name: str, **kwargs):
            """Consume RPC events for the given API"""
            raise NotImplementedError("Not yet supported by mocks")

    return TestEventTransport


def make_test_schema_transport():
    class TestSchemaTransport(SchemaTransport):
        def __init__(self):
            self.schemas = {}

        async def store(self, api_name: str, schema: Dict, ttl_seconds: int):
            self.schemas[api_name] = schema

        async def ping(self, api_name: str, schema: Dict, ttl_seconds: int):
            pass

        async def load(self) -> Dict[str, Dict]:
            return self.schemas

    return TestSchemaTransport


# Command mocking
# These tools are not part of the public API, but are used for internal testing,
# and to power the above BusMocker (which is part of the public API)


class CommandList(list):
    def types(self):
        return [type(i for i in self)]

    def get_all(self, type_: Type[T]) -> List[T]:
        commands = []
        for i in self:
            if type(i) == type_:
                commands.append(i)
        return commands

    def get(self, type_: Type[T], multiple_ok=False) -> T:
        commands = self.get_all(type_)
        if not commands:
            raise ValueError(f"No command found of type {type_}")
        elif len(commands) > 1 and not multiple_ok:
            raise ValueError(f"Multiple ({len(commands)}) commands found of type {type_}")
        else:
            return commands[0]

    def has(self, type_: Type[T]):
        return any(type(i) == type_ for i in self)

    def count(self, type_: Type[T]):
        len(self.get_all(type_))


Command = TypeVar("Command", bound=NamedTuple)


class QueueMockContext:
    def __init__(self, queue: InternalQueue):
        self.queue = queue
        self.put_items: List[Tuple[Command, asyncio.Event]] = []
        self.got_items: List[Tuple[Command, asyncio.Event]] = []

        self._patched_put_nowait: Optional[_patch] = None
        self._patched_get_nowait: Optional[_patch] = None

        self._blackhole: Set[Type[Command]] = set()

    def __enter__(self) -> "QueueMockContext":
        self.put_items = []
        self.got_items = []

        self._orig_put_nowait = self.queue.put_nowait
        self._orig_get_nowait = self.queue.get_nowait

        self._patched_put_nowait = patch.object(self.queue, "put_nowait", wraps=self._put_nowait)
        self._patched_get_nowait = patch.object(self.queue, "get_nowait", wraps=self._get_nowait)

        self._patched_put_nowait.start()
        self._patched_get_nowait.start()

        return self

    def blackhole(self, *command_types):
        self._blackhole = self._blackhole.union(set(command_types))

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._patched_put_nowait.stop()
        self._patched_get_nowait.stop()

    def _put_nowait(self, item, *args, **kwargs):
        self.put_items.append(item)

        try:
            command, event = item
        except ValueError:
            # The item being added isn't a command
            pass
        else:
            # Don't call blackholed commands
            if type(command) in self._blackhole:
                return

        return self._orig_put_nowait(item, *args, **kwargs)

    def _get_nowait(self, *args, **kwargs):
        item = self._orig_get_nowait(*args, **kwargs)
        self.got_items.append(item)
        return item

    @property
    def put_commands(self) -> CommandList:
        return CommandList([i[0] for i in self.put_items])

    @property
    def got_commands(self) -> CommandList:
        return CommandList([i[0] for i in self.got_items])

    @property
    def items(self):
        # Just a little shortcut for the common use case
        return self.put_items

    @property
    def commands(self):
        # Just a little shortcut for the common use case
        return self.put_commands


class BusQueueMockerContext:
    class Queues(NamedTuple):
        to_transport: QueueMockContext
        from_transport: QueueMockContext

    def __init__(self, client: BusClient):
        self.client = client
        self.event: Optional[BusQueueMockerContext.Queues] = None
        self.rpc_result: Optional[BusQueueMockerContext.Queues] = None
        self.errors: Optional[QueueMockContext] = None

    def __enter__(self):
        self.event = BusQueueMockerContext.Queues(
            to_transport=QueueMockContext(self.client.event_client.producer.queue).__enter__(),
            from_transport=QueueMockContext(self.client.event_client.consumer.queue).__enter__(),
        )
        self.rpc_result = BusQueueMockerContext.Queues(
            to_transport=QueueMockContext(self.client.rpc_result_client.producer.queue).__enter__(),
            from_transport=QueueMockContext(
                self.client.rpc_result_client.consumer.queue
            ).__enter__(),
        )
        self.errors = QueueMockContext(self.client.error_queue).__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.event.to_transport.__exit__(exc_type, exc_val, exc_tb)
        self.event.from_transport.__exit__(exc_type, exc_val, exc_tb)
        self.rpc_result.to_transport.__exit__(exc_type, exc_val, exc_tb)
        self.rpc_result.from_transport.__exit__(exc_type, exc_val, exc_tb)



class LightbusException(Exception):
    pass


class InvalidMessage(LightbusException):
    pass


class InvalidBusPathConfiguration(LightbusException):
    pass


class InvalidSerializerConfiguration(LightbusException):
    pass


class UnsupportedUse(LightbusException):
    pass


class UnknownApi(LightbusException):
    pass


class InvalidApiRegistryEntry(LightbusException):
    pass


class InvalidApiForSchemaCreation(LightbusException):
    pass


class InvalidApiEventConfiguration(LightbusException):
    pass


class MisconfiguredApiOptions(LightbusException):
    pass


class EventNotFound(LightbusException):
    pass


class InvalidEventArguments(LightbusException):
    pass


class InvalidEventListener(LightbusException):
    pass


class NothingToListenFor(LightbusException):
    pass


class CannotBlockHere(LightbusException):
    pass


class LightbusTimeout(LightbusException):
    pass


class SuddenDeathException(LightbusException):
    """Used to kill an invocation for testing purposes"""

    pass


class PluginsNotLoaded(LightbusException):
    pass


class InvalidPlugins(LightbusException):
    pass


class PluginHookNotFound(LightbusException):
    pass


class LightbusWorkerError(LightbusException):
    pass


class LightbusShutdownInProgress(LightbusException):
    pass


class InvalidSchema(LightbusException):
    pass


class SchemaNotFound(LightbusException):
    pass


class NoResponseSchemaFound(SchemaNotFound):
    pass


class TransportNotFound(LightbusException):
    pass


class TooManyTransportsForApi(LightbusException):
    pass


class NoTransportConfigured(LightbusException):
    pass


class NoApisToListenOn(LightbusException):
    pass


class InvalidName(LightbusException):
    pass


class InvalidParameters(LightbusException):
    pass


class ApisMustUseSameTransport(LightbusException):
    pass


class OnlyAvailableOnRootNode(LightbusException):
    pass


class UnexpectedConfigurationFormat(LightbusException):
    pass


class UnsupportedOptionValue(LightbusException):
    pass


class DatabaseConnectionRequired(LightbusException):
    pass


class DuplicateMessage(LightbusException):
    pass


class NoApisSpecified(LightbusException):
    pass


class FailedToImportBusModule(LightbusException):
    pass


class NoBusFoundInBusModule(LightbusException):
    pass


class TransportIsClosed(LightbusException):
    pass


class DeformError(LightbusException):
    pass


class ValidationError(LightbusException):
    pass


class InvalidSchedule(LightbusException):
    pass


class TransportsNotInstalled(LightbusException):
    pass


class MustRunInBusThread(LightbusException):
    pass


class MustNotRunInBusThread(LightbusException):
    pass


class LightbusExit(LightbusException):
    pass


class BusAlreadyClosed(LightbusException):
    pass


class LightbusServerMustStartInMainThread(LightbusException):
    pass


class RemoteSchemasNotLoaded(LightbusException):
    pass


class TransportPoolIsClosed(LightbusException):
    pass


class CannotShrinkEmptyPool(LightbusException):
    pass


class PoolProxyError(LightbusException):
    pass


class CannotProxyPrivateMethod(PoolProxyError):
    pass


class CannotProxyProperty(PoolProxyError):
    pass


class CannotProxySynchronousMethod(PoolProxyError):
    pass


class AsyncFunctionOrMethodRequired(LightbusException):
    pass


class ListenersAlreadyStarted(LightbusException):
    pass


class DuplicateListenerName(LightbusException):
    pass



import argparse
import logging
import sys
from inspect import isclass

import lightbus
from lightbus.commands import utilities as command_utilities
from lightbus.plugins import PluginRegistry
from lightbus.utilities.async_tools import block

logger = logging.getLogger(__name__)


class Command:
    def setup(self, parser, subparsers):
        parser_shell = subparsers.add_parser(
            "shell",
            help="Provide an interactive Lightbus shell",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        command_utilities.setup_common_arguments(parser_shell)
        parser_shell.set_defaults(func=self.handle)

    def handle(self, args, config, plugin_registry: PluginRegistry, fake_it=False):
        command_utilities.setup_logging(args.log_level or "warning", config)

        try:
            # pylint: disable=unused-import,cyclic-import,import-outside-toplevel
            import bpython
            from bpython.curtsies import main as bpython_main
        except ImportError:  # pragma: no cover
            print("Lightbus shell requires bpython. Run `pip install bpython` to install bpython.")
            sys.exit(1)
            return  # noqa

        lightbus_logger = logging.getLogger("lightbus")
        lightbus_logger.setLevel(logging.WARNING)

        bus_module, bus = command_utilities.import_bus(args)
        block(bus.client.lazy_load_now())

        objects = {k: v for k, v in lightbus.__dict__.items() if isclass(v)}
        objects.update(bus=bus)

        block(plugin_registry.execute_hook("receive_args", args=args), timeout=5)

        # Ability to not start up the repl is useful for testing
        if not fake_it:
            bpython_main(
                args=["-i", "-q"],
                locals_=objects,
                welcome_message="Welcome to the Lightbus shell. Use `bus` to access your bus.",
            )



import argparse
import logging
from pathlib import Path

import pkg_resources


logger = logging.getLogger(__name__)


class Command:
    def setup(self, parser, subparsers):
        parser_version = subparsers.add_parser(
            "version", help="Show the currently installed Lightbus version"
        )
        # Read version directly out of pyproject.toml. Useful for the release process
        parser_version.add_argument("--pyproject", action="store_true", help=argparse.SUPPRESS)
        # Show the version to be used for creating the docs
        parser_version.add_argument("--docs", action="store_true", help=argparse.SUPPRESS)
        parser_version.set_defaults(func=self.handle)

    def handle(self, args):
        if args.pyproject:
            import lightbus
            import toml

            file_path = Path(lightbus.__file__).parent.parent / "pyproject.toml"
            with file_path.open() as f:
                version = toml.load(f)["tool"]["poetry"]["version"]
            if args.docs:
                version = ".".join(version.split(".")[:2])
            print(version)
        else:
            print(pkg_resources.get_distribution("lightbus").version)



import threading
from inspect import iscoroutinefunction, isasyncgenfunction
from typing import NamedTuple, List, TypeVar, Type, Generic, TYPE_CHECKING

from lightbus.exceptions import (
    TransportPoolIsClosed,
    CannotShrinkEmptyPool,
    CannotProxySynchronousMethod,
    CannotProxyPrivateMethod,
    CannotProxyProperty,
)

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from lightbus.config import Config
    from lightbus.transports.base import Transport

    VT = TypeVar("VT", bound=Transport)
else:
    VT = TypeVar("VT")


class TransportPool(Generic[VT]):
    """ Pool for managing access to transports

    This pool with function as a transparent proxy to the underlying transports.
    In most cases you shouldn't need to access the underlying transports. If you
    do you can use the context manage as follows:

        async with transport_pool as transport:
            transport.send_event(...)

    Note that this pool will only perform pooling within the thread in which the
    pool was created. If another thread uses the pool then the pool will be bypassed.
    In this case, a new transport will always be created on checkout, and this
    transport will then be immediately closed when checked back in.

    This is because the pool will normally be closed sometime after the thread has
    completed, at which point each transport in the pool will be closed. However, closing
    the transport requires access to the event loop for the specific transport, but that
    loop would have been closed when the thread shutdown. It therefore becomes impossible to
    the transport cleanly. Therefore, in the case of threads, we create new transports on
    checkout, and close and discard the transport on checkin.

    This will have some performance impact for non-async user-provided-callables which need to
    access the bus. These callables area run in a thread, and so will need fresh connections.
    """

    def __init__(self, transport_class: Type[VT], transport_config: NamedTuple, config: "Config"):
        self.transport_class = transport_class
        self.transport_config = transport_config
        self.config = config
        self.closed = False

        self.lock = threading.RLock()
        self.pool: List[VT] = []
        self.checked_out = set()
        self.context_stack: List[VT] = []
        self.home_thread = threading.current_thread()

    def __repr__(self):
        return f"<Pool of {self.transport_class.__name__} at 0x{id(self):02x}>"

    def __hash__(self):
        return hash((self.transport_class, self.transport_config))

    def __eq__(self, other):
        return hash(self) == hash(other)

    async def grow(self):
        with self.lock:
            new_transport = await self._create_transport()
            self.pool.append(new_transport)

    async def shrink(self):
        with self.lock:
            try:
                old_transport = self.pool.pop(0)
            except IndexError:
                raise CannotShrinkEmptyPool(
                    "Transport pool is already empty, cannot shrink it further"
                )
            await self._close_transport(old_transport)

    async def checkout(self) -> VT:
        if self.closed:
            raise TransportPoolIsClosed("Cannot get a connection, transport pool is closed")

        if threading.current_thread() != self.home_thread:
            return await self._create_transport()
        else:
            with self.lock:
                if not self.pool:
                    await self.grow()

                transport = self.pool.pop(0)
                self.checked_out.add(transport)
            return transport

    async def checkin(self, transport: VT):
        if threading.current_thread() != self.home_thread:
            return await self._close_transport(transport)
        else:
            with self.lock:
                self.checked_out.discard(transport)
                self.pool.append(transport)
                if self.closed:
                    await self._close_all()

    @property
    def free(self) -> int:
        return len(self.pool)

    @property
    def in_use(self) -> int:
        return len(self.checked_out)

    @property
    def total(self) -> int:
        return self.free + self.in_use

    async def __aenter__(self) -> VT:
        transport = await self.checkout()
        self.context_stack.append(transport)
        return transport

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        transport = self.context_stack.pop()
        await self.checkin(transport)

    async def close(self):
        with self.lock:
            self.closed = True
            await self._close_all()

    async def _close_all(self):
        with self.lock:
            while self.pool:
                await self._close_transport(self.pool.pop())

    async def _create_transport(self) -> VT:
        new_transport = self.transport_class.from_config(
            config=self.config, **self.transport_config._asdict()
        )
        await new_transport.open()
        return new_transport

    async def _close_transport(self, transport: VT):
        """Close a specific transport"""
        await transport.close()

    def __getattr__(self, item):
        async def fn_pool_wrapper(*args, **kwargs):
            async with self as transport:
                return await getattr(transport, item)(*args, **kwargs)

        async def gen_pool_wrapper(*args, **kwargs):
            async with self as transport:
                async for value in getattr(transport, item)(*args, **kwargs):
                    yield value

        attr = getattr(self.transport_class, item, None)

        if not attr:
            raise AttributeError(
                f"Neither the transport pool {repr(self)} nor the transport with class "
                f"{repr(self.transport_class)} has an attribute named {item}"
            )
        elif item[0] == "_":
            raise CannotProxyPrivateMethod(
                f"Cannot proxy private method calls to transport. Use the pool's async context or "
                f"checkout() method if you really need to access private methods. (Private methods "
                f"are ones whose name starts with an underscore)"
            )
        elif not callable(attr):
            raise CannotProxyProperty(
                f"Cannot proxy property access on transports. Use the pool's async context or "
                f"checkout() method to get access to a transport directly."
            )
        else:
            if iscoroutinefunction(attr):
                return fn_pool_wrapper
            elif isasyncgenfunction(attr):
                return gen_pool_wrapper
            else:
                raise CannotProxySynchronousMethod(
                    f"{self.transport_class.__name__}.{item}() is synchronous "
                    f"and must be accessed directly and not via the pool"
                )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lightbus',
 'lightbus.client',
 'lightbus.client.docks',
 'lightbus.client.internal_messaging',
 'lightbus.client.subclients',
 'lightbus.commands',
 'lightbus.config',
 'lightbus.plugins',
 'lightbus.schema',
 'lightbus.serializers',
 'lightbus.transports',
 'lightbus.transports.redis',
 'lightbus.utilities',
 'lightbus_vendored',
 'lightbus_vendored.jsonpath']

package_data = \
{'': ['*']}

install_requires = \
['aioredis>=1.2.0', 'jsonschema>=3.2', 'pyyaml>=3.12']

entry_points = \
{'console_scripts': ['lightbus = lightbus.commands:lightbus_entry_point'],
 'lightbus_event_transports': ['debug = lightbus:DebugEventTransport',
                               'redis = lightbus:RedisEventTransport'],
 'lightbus_plugins': ['internal_metrics = '
                      'lightbus.plugins.metrics:MetricsPlugin',
                      'internal_state = lightbus.plugins.state:StatePlugin'],
 'lightbus_result_transports': ['debug = lightbus:DebugResultTransport',
                                'redis = lightbus:RedisResultTransport'],
 'lightbus_rpc_transports': ['debug = lightbus:DebugRpcTransport',
                             'redis = lightbus:RedisRpcTransport'],
 'lightbus_schema_transports': ['debug = lightbus:DebugSchemaTransport',
                                'redis = lightbus:RedisSchemaTransport']}

setup_kwargs = {
    'name': 'lightbus',
    'version': '0.11.0a1',
    'description': 'RPC & event framework for Python 3',
    'long_description': "\nWhat is Lightbus?\n=================\n\n\n.. image:: https://img.shields.io/circleci/build/github/adamcharnock/lightbus\n   :target: https://circleci.com/gh/adamcharnock/lightbus/tree/master\n   :alt: CircleCI\n\n\n.. image:: https://api.codacy.com/project/badge/Grade/801d031fd2714b4f9c643182f1fbbd0b\n   :target: https://www.codacy.com/app/adamcharnock/lightbus?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=adamcharnock/lightbus&amp;utm_campaign=Badge_Grade\n   :alt: Codacy Badge\n\n\n.. image:: https://api.codacy.com/project/badge/Coverage/801d031fd2714b4f9c643182f1fbbd0b\n   :target: https://www.codacy.com/app/adamcharnock/lightbus?utm_source=github.com&utm_medium=referral&utm_content=adamcharnock/lightbus&utm_campaign=Badge_Coverage\n   :alt: Codacy Badge\n\n\n.. image:: https://img.shields.io/discord/645218336229031946\n   :target: https://discord.gg/2j594ws\n   :alt: Discord\n\n\n.. image:: https://img.shields.io/badge/Contributor%20Covenant-v2.0%20adopted-ff69b4.svg\n   :target: https://lightbus.org/reference/code-of-conduct/\n   :alt: Contributor Covenant\n\n\nLightbus allows your backend processes to communicate, run background tasks,\nand expose internal APIs.\n\nLightbus uses Redis as its underlying transport, although support\nfor other platforms may eventually be added.\n\nLightbus requires Python 3.7 or above.\n\n**Full documentation can be found at https://lightbus.org**\n\nDesigned for ease of use\n------------------------\n\nLightbus is designed with developers in mind. The syntax aims to\nbe intuitive and familiar, and common problems are caught with\nclear and helpful error messages.\n\nFor example, a naïve authentication API:\n\n.. code-block:: python3\n\n   class AuthApi(Api):\n       user_registered = Event(parameters=('username', 'email'))\n\n       class Meta:\n           name = 'auth'\n\n       def check_password(self, user, password):\n           return (\n               user == 'admin'\n               and password == 'secret'\n           )\n\nThis can be called as follows:\n\n.. code-block:: python3\n\n   import lightbus\n\n   bus = lightbus.create()\n\n   bus.auth.check_password(\n       user='admin',\n       password='secret'\n   )\n   # Returns true\n\nYou can also listen for events:\n\n.. code-block:: python3\n\n   import lightbus\n\n   bus = lightbus.create()\n\n   def send_signup_email(event_message,\n                         username, email):\n       send_mail(email,\n           subject=f'Welcome {username}'\n       )\n\n   @bus.client.on_start()\n   def bus_start():\n       bus.auth.user_registered.listen(\n           send_signup_email\n       )\n\n**To get started checkout the documentation at https://lightbus.org.**\n",
    'author': 'Adam Charnock',
    'author_email': 'adam@adamcharnock.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://lightbus.org',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



from typing import Dict

from lightbus.exceptions import (
    UnknownApi,
    InvalidApiRegistryEntry,
    EventNotFound,
    MisconfiguredApiOptions,
    InvalidApiEventConfiguration,
)


__all__ = ["Api", "Event"]


class ApiRegistry:
    def __init__(self):
        self._apis: Dict[str, Api] = dict()

    def add(self, api: "Api"):
        if isinstance(api, type):
            raise InvalidApiRegistryEntry(
                "An attempt was made to add a type to the API registry. This "
                "is probably because you are trying to add the API class, rather "
                "than an instance of the API class.\n"
                "\n"
                "Use bus.client.register_api(MyApi()), rather than bus.client.register_api(MyApi)"
            )

        self._apis[api.meta.name] = api

    def get(self, name) -> "Api":
        try:
            return self._apis[name]
        except KeyError:
            raise UnknownApi(
                "An API named '{}' was requested from the registry but the "
                "registry does not recognise it. Maybe the incorrect API name "
                "was specified, or maybe the API has not been registered.".format(name)
            )

    def remove(self, name) -> None:
        try:
            del self._apis[name]
        except KeyError:
            raise UnknownApi(
                "An attempt was made to remove an API named '{}' from the registry, but the API "
                "could not be found. Maybe the incorrect API name "
                "was specified, or maybe the API has not been registered.".format(name)
            )

    def public(self):
        return [api for api in self._apis.values() if not api.meta.internal]

    def internal(self):
        return [api for api in self._apis.values() if api.meta.internal]

    def all(self):
        return list(self._apis.values())

    def names(self):
        return list(self._apis.keys())


class ApiOptions:
    name: str
    internal: bool = False
    version: int = 1

    def __init__(self, options):
        for k, v in options.items():
            if not k.startswith("_"):
                setattr(self, k, v)


class ApiMetaclass(type):
    """ API Metaclass

    Validates options in the API's Meta class and populates the
    API class' `meta` attribute.
    """

    def __init__(cls, name, bases=None, dict_=None):
        is_api_base_class = name == "Api" and not bases
        if is_api_base_class:
            super(ApiMetaclass, cls).__init__(name, bases, dict_)
        else:
            options = dict_.get("Meta", None)
            if options is None:
                raise MisconfiguredApiOptions(
                    f"API class {name} does not contain a class named 'Meta'. Each API definition "
                    f"must contain a child class named 'Meta' which can contain configurations options. "
                    f"For example, the 'name' option is required and specifies "
                    f"the name used to access the API on the bus."
                )
            cls.sanity_check_options(name, options)
            cls.meta = ApiOptions(cls.Meta.__dict__.copy())
            super(ApiMetaclass, cls).__init__(name, bases, dict_)

            if cls.meta.name == "default" or cls.meta.name.startswith("default."):
                raise MisconfiguredApiOptions(
                    f"API class {name} is named 'default', or starts with 'default.'. "
                    f"This is a reserved name and is not allowed, please change it to something else."
                )

    def sanity_check_options(cls, name, options):
        if not getattr(options, "name", None):
            raise MisconfiguredApiOptions(
                "API class {} does not specify a name option with its "
                "'Meta' options."
                "".format(name)
            )


class Api(metaclass=ApiMetaclass):
    class Meta:
        name = None

    def get_event(self, name) -> "Event":
        event = getattr(self, name, None)
        if isinstance(event, Event):
            return event
        else:
            raise EventNotFound("Event named {}.{} could not be found".format(self, name))

    def __str__(self):
        return self.meta.name


class Event:
    def __init__(self, parameters=tuple()):
        # Ensure you update the __copy__() method if adding other instance variables below
        if isinstance(parameters, str):
            raise InvalidApiEventConfiguration(
                f"You appear to have passed a string value of {repr(parameters)} "
                f"for your API's event's parameters. This should be a list or a tuple, "
                f"not a string. You probably missed a comma when defining your "
                f"tuple of parameter names."
            )
        self.parameters = parameters



import inspect
import itertools
import logging
from decimal import Decimal
from typing import Union, Any, Tuple, Sequence, Mapping, Callable, TYPE_CHECKING

import datetime
from enum import Enum
from uuid import UUID

from lightbus.utilities.deforming import deform_to_bus
from lightbus.utilities.type_checks import parse_hint, issubclass_safe, get_property_default

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from lightbus.api import Event

NoneType = type(None)
empty = inspect.Parameter.empty
logger = logging.getLogger(__name__)


SCHEMA_URI = "http://json-schema.org/draft-07/schema#"


def make_rpc_parameter_schema(api_name, method_name, method):
    """Create a full parameter JSON schema for the given RPC
    """
    parameters = inspect.signature(method).parameters.values()
    schema = make_parameter_schema(parameters)
    schema["title"] = "RPC {}.{}() parameters".format(api_name, method_name)
    return schema


def make_response_schema(api_name: str, method_name: str, method: Callable):
    """Create a full response JSON schema for the given RPC
    """
    sig = inspect.signature(method)
    schema = return_type_to_schema(sig.return_annotation)
    schema["title"] = "RPC {}.{}() response".format(api_name, method_name)
    schema["$schema"] = SCHEMA_URI
    return schema


def make_event_parameter_schema(api_name, method_name, event: "Event"):
    """Create a full parameter JSON schema for the given event
    """
    parameters = _normalise_event_parameters(event.parameters)
    schema = make_parameter_schema(parameters)
    schema["title"] = "Event {}.{} parameters".format(api_name, method_name)
    return schema


def make_parameter_schema(parameters: Sequence[inspect.Parameter]):
    """Create a full JSON schema for the given parameters
    """
    parameter_schema = {
        "$schema": SCHEMA_URI,
        "type": "object",
        "additionalProperties": False,
        "properties": {},
        "required": [],
    }

    for parameter in parameters:
        if parameter.kind in (parameter.POSITIONAL_ONLY, parameter.VAR_POSITIONAL):
            logger.warning(
                "Positional-only arguments are not supported in event or RPC parameters: {}".format(
                    parameter
                )
            )
            continue

        if parameter.kind == parameter.VAR_KEYWORD:
            parameter_schema["additionalProperties"] = True
            continue

        parameter_schema["properties"][parameter.name] = parameter_to_schema(parameter)
        if parameter.default is empty:
            parameter_schema["required"].append(parameter.name)

    # required key should not be present if it is empty
    if not parameter_schema["required"]:
        parameter_schema.pop("required")

    return parameter_schema


def parameter_to_schema(parameter):
    """Convert an `inspect.Parameter` to a single JSON schema

    This will create the schemas using `annotation_to_json_schemas()`,
    set the default values based on the parameters default value
    annotation, and combine the schemas with `wrap_with_any_of()`
    """
    type_schemas = annotation_to_json_schemas(parameter.annotation, parameter.default)
    if not type_schemas:
        return {}

    schemas = []
    for type_schema in type_schemas:
        if parameter.default is not empty:
            type_schema["default"] = parameter.default
        schemas.append(type_schema)

    return wrap_with_any_of(schemas)


def return_type_to_schema(type_):
    """Convert the return type annotation into a json schema

    Mostly a wrapper around `python_type_to_json_schemas()`.
    """
    if type_ is empty:
        return {}

    type_schemas = python_type_to_json_schemas(type_)
    if not type_schemas:
        return {}

    return wrap_with_any_of(type_schemas)


def annotation_to_json_schemas(annotation, default=empty):
    """Convert an `inspect.Parameter` to JSON schemas

    This wraps `python_type_to_json_schemas()` with a little extra logic. Specifically,
    it determines the python type from:

        1. The parameter's annotation, falling back to...
        2. The type of the parameters default value

    This will return a list of JSON schemas. See note one `python_type_to_json_schemas()`
    """
    if annotation is not empty:
        schemas = python_type_to_json_schemas(annotation)

        # Allow nulls if the parameter has a default value of None
        if default is not empty and default is None:
            null_schema = {"type": "null"}
            if null_schema not in schemas:
                schemas.append(null_schema)

    elif default is not empty and default is not None:
        schemas = python_type_to_json_schemas(type(default))
    else:
        schemas = [{}]

    return schemas


def python_type_to_json_schemas(type_):
    """Convert a python type hint to its JSON schema representations

    Note that a type hint may actually have several possible representations,
    which is why this function returns a list of schemas. An example of this is
    the `Union` type hint. These are later combined via `wrap_with_any_of()`
    """
    # pylint: disable=too-many-return-statements
    type_, hint_args = parse_hint(type_)

    if type_ == Union:
        return list(itertools.chain(*map(python_type_to_json_schemas, hint_args)))
    if type_ == empty:
        return [{}]
    elif type_ in (Any, ...):
        return [{}]
    elif issubclass_safe(type_, (str, bytes, complex, UUID)):
        return [{"type": "string"}]
    elif issubclass_safe(type_, Decimal):
        return [{"type": "string", "pattern": r"^-?\d+(\.\d+)?$"}]
    elif issubclass_safe(type_, (bool,)):
        return [{"type": "boolean"}]
    elif issubclass_safe(type_, (float,)):
        return [{"type": "number"}]
    elif issubclass_safe(type_, (int,)):
        return [{"type": "integer"}]
    elif issubclass_safe(type_, (Mapping,)) and hint_args and hint_args[0] == str:
        # Mapping with strings as keys
        return [
            {
                "type": "object",
                "additionalProperties": wrap_with_any_of(python_type_to_json_schemas(hint_args[1])),
            }
        ]
    elif issubclass_safe(type_, (dict, Mapping)):
        return [{"type": "object"}]
    elif issubclass_safe(type_, tuple) and hasattr(type_, "_fields"):
        # Named tuple
        return [make_custom_object_schema(type_, property_names=type_._fields)]
    elif issubclass_safe(type_, Enum) and type_.__members__:
        # Enum
        enum_first_value = list(type_.__members__.values())[0].value
        schema = {}
        try:
            schema["type"] = python_type_to_json_schemas(type(enum_first_value))[0]["type"]
            schema["enum"] = [v.value for v in type_.__members__.values()]
        except KeyError:
            logger.warning(f"Could not determine type for values in enum: {type_}")
        return [schema]
    elif issubclass_safe(type_, (Tuple,)) and hint_args:
        return [
            {
                "type": "array",
                "maxItems": len(hint_args),
                "minItems": len(hint_args),
                "items": [
                    wrap_with_any_of(python_type_to_json_schemas(sub_type))
                    for sub_type in hint_args
                ],
            }
        ]
    elif issubclass_safe(type_, (list, tuple, set)):
        schema = {"type": "array"}
        if hint_args:
            schema["items"] = wrap_with_any_of(python_type_to_json_schemas(hint_args[0]))
        return [schema]
    elif issubclass_safe(type_, NoneType) or type_ is None:
        return [{"type": "null"}]
    elif issubclass_safe(type_, (datetime.datetime)):
        return [{"type": "string", "format": "date-time"}]
    elif issubclass_safe(type_, (datetime.date)):
        return [{"type": "string", "format": "date"}]
    elif issubclass_safe(type_, (datetime.time)):
        return [{"type": "string", "format": "time"}]
    elif getattr(type_, "__annotations__", None):
        # Custom class
        return [make_custom_object_schema(type_)]
    else:
        logger.warning(
            f"Could not convert python type to json schema type: {type_}. If it is a class, "
            f"ensure it's class-level variables have type hints."
        )
        return [{}]


def make_custom_object_schema(type_, property_names=None):
    """Convert a named tuple into a JSON schema

    While a lot of conversations happen in `python_type_to_json_schemas()`, conversion of
    annotated classes is a little more involved. It therefore gets its own function.

    While this is mainly used for named tuples within lightbus, it can also be used for
    general classes.
    """
    if property_names is None:
        property_names = [
            p for p in set(list(type_.__annotations__.keys()) + dir(type_)) if p[0] != "_"
        ]

    properties = {}
    required = []
    for property_name in property_names:
        default = empty

        if callable(getattr(type_, property_name, None)):
            # is a method
            continue

        default = get_property_default(type_, property_name)

        if hasattr(type_, "__annotations__"):
            properties[property_name] = wrap_with_any_of(
                annotation_to_json_schemas(
                    annotation=type_.__annotations__.get(property_name, None), default=default
                )
            )
        elif default is not empty:
            properties[property_name] = wrap_with_any_of(python_type_to_json_schemas(type(default)))
        else:
            properties[property_name] = {}

        if default is empty:
            required.append(property_name)
        else:
            properties[property_name]["default"] = deform_to_bus(default)

    schema = {
        "type": "object",
        "title": type_.__name__,
        "properties": properties,
        "required": required,
        "additionalProperties": False,
    }

    # required key should not be present if it is empty
    if not schema["required"]:
        schema.pop("required")

    return schema


def wrap_with_any_of(schemas: Sequence):
    """Take multiple JSON schemas and combine them using the `anyOf` JSON schema notation"""
    if len(schemas) == 1:
        return schemas[0]
    else:
        return {"anyOf": schemas}


def _normalise_event_parameters(parameters: Sequence) -> Sequence:
    """Ensure the given event parameters are in the long format

    Event parameters can be specified as strings or `inspect.Parameter`
    instances. Ensure all provided parameters are all in the latter form.
    """
    # pylint: disable=cyclic-import,import-outside-toplevel
    from lightbus.schema import Parameter

    normalised = []
    for parameter in parameters:
        if isinstance(parameter, str):
            normalised.append(Parameter(name=parameter))
        else:
            normalised.append(parameter)
    return normalised
